import pytest
import numpy as np
import scipy.stats as sts

from virocon import (
    WeibullDistribution,
    LogNormalDistribution,
    NormalDistribution,
    ExponentiatedWeibullDistribution,
    GeneralizedGammaDistribution,
    ScipyDistribution,
)


def test_scipydistribution():
    # Generalized Gamma Distribution
    generalized_gamma_name = "gengamma"
    generalized_gamma_dist = sts.gengamma

    sample_params = {"a": 1.5, "c": 2, "loc": 0, "scale": 0.5}
    n = 1000
    sample = generalized_gamma_dist.rvs(**sample_params, size=n, random_state=42)
    sample.sort()
    x = np.linspace(sample.min(), sample.max())
    p = np.linspace(0, 1, num=n)

    class GeneralizedGammaDistributionByName(ScipyDistribution):
        scipy_dist_name = generalized_gamma_name

    class GeneralizedGammaDistributionByDist(ScipyDistribution):
        scipy_dist = generalized_gamma_dist

    # test wrong parameter instantiation
    with pytest.raises(TypeError):
        gengamma_by_name = GeneralizedGammaDistributionByName(floc=0)
    with pytest.raises(TypeError):
        gengamma_by_dist = GeneralizedGammaDistributionByDist(floc=0)

    gengamma_by_name = GeneralizedGammaDistributionByName(f_loc=0)
    gengamma_by_dist = GeneralizedGammaDistributionByDist(f_loc=0)

    # fit
    gengamma_by_name.fit(sample)
    gengamma_by_dist.fit(sample)

    assert list(gengamma_by_name.parameters.keys()) == ["a", "c", "loc", "scale"]
    assert list(gengamma_by_dist.parameters.keys()) == ["a", "c", "loc", "scale"]

    np.testing.assert_allclose(
        gengamma_by_name.parameters["a"], sample_params["a"], rtol=0.1
    )
    np.testing.assert_allclose(
        gengamma_by_name.parameters["c"], sample_params["c"], rtol=0.1
    )
    assert gengamma_by_name.parameters["loc"] == 0.0
    np.testing.assert_allclose(
        gengamma_by_name.parameters["scale"], sample_params["scale"], rtol=0.1
    )

    np.testing.assert_allclose(
        gengamma_by_name.parameters["a"], sample_params["a"], rtol=0.1
    )
    np.testing.assert_allclose(
        gengamma_by_name.parameters["c"], sample_params["c"], rtol=0.1
    )
    assert gengamma_by_name.parameters["loc"] == 0.0
    np.testing.assert_allclose(
        gengamma_by_name.parameters["scale"], sample_params["scale"], rtol=0.1
    )

    # sample
    sample_mean = sample.mean()
    sample_var = sample.var()

    sample_by_name = gengamma_by_name.draw_sample(n, random_state=42)
    sample_by_dist = gengamma_by_dist.draw_sample(n, random_state=42)
    mean_by_name = sample_by_name.mean()
    var_by_name = sample_by_name.var()
    mean_by_dist = sample_by_dist.mean()
    var_by_dist = sample_by_dist.var()
    assert mean_by_name == mean_by_dist
    assert var_by_name == var_by_dist

    np.testing.assert_allclose(sample_mean, mean_by_name, rtol=0.1)
    np.testing.assert_allclose(sample_var, var_by_name, rtol=0.1)

    # pdf
    pdf_by_name = gengamma_by_name.pdf(x)
    pdf_by_dist = gengamma_by_dist.pdf(x)
    assert (pdf_by_name == pdf_by_dist).all()
    assert (
        generalized_gamma_dist.pdf(x, **gengamma_by_name.parameters) == pdf_by_name
    ).all()

    # cdf
    cdf_by_name = gengamma_by_name.cdf(x)
    cdf_by_dist = gengamma_by_dist.cdf(x)
    assert (cdf_by_name == cdf_by_dist).all()
    assert (
        generalized_gamma_dist.cdf(x, **gengamma_by_name.parameters) == cdf_by_name
    ).all()

    # icdf
    icdf_by_name = gengamma_by_name.icdf(p)
    icdf_by_dist = gengamma_by_dist.icdf(p)
    assert (icdf_by_name == icdf_by_dist).all()
    assert (
        generalized_gamma_dist.ppf(p, **gengamma_by_name.parameters) == icdf_by_name
    ).all()


def test_exponentiated_weibull_distribution_cdf():
    """
    Tests the CDF of the exponentiated Weibull distribution.
    """
    # Define dist with parameters from the distribution fitted to
    # dataset A with the MLE in https://arxiv.org/pdf/1911.12835.pdf .
    dist = ExponentiatedWeibullDistribution(alpha=0.0373, beta=0.4743, delta=46.6078)

    # CDF(1) should be roughly 0.7, see Figure 12 in
    # https://arxiv.org/pdf/1911.12835.pdf .
    p = dist.cdf(1)
    np.testing.assert_allclose(p, 0.7, atol=0.1)

    # CDF(4) should be roughly 0.993, see Figure 12 in
    # https://arxiv.org/pdf/1911.12835.pdf .
    p = dist.cdf(4)
    assert p > 0.99
    assert p < 0.999

    # CDF(negative value) should be 0
    p = dist.cdf(-1)
    assert p == 0


def test_exponentiated_weibull_distribution_icdf():
    """
    Tests the ICDF of the exponentiated Weibull distribution.
    """
    # Define dist with parameters from the distribution fitted to
    # dataset A with the MLE in https://arxiv.org/pdf/1911.12835.pdf .
    dist = ExponentiatedWeibullDistribution(alpha=0.0373, beta=0.4743, delta=46.6078)

    # ICDF(0.5) should be roughly 0.8, see Figure 12 in
    # https://arxiv.org/pdf/1911.12835.pdf .
    x = dist.icdf(0.5)
    assert x > 0.5
    assert x < 1

    # ICDF(0.9) should be roughly 1.8, see Figure 12
    # in https://arxiv.org/pdf/1911.12835.pdf .
    x = dist.icdf(0.9)
    assert x > 1
    assert x < 2

    # ICDF(value greater than 1) should be nan.
    x = dist.icdf(5)
    assert np.isnan(x)


def test_exponentiated_weibull_distribution_pdf():
    """
    Tests the PDF of the exponentiated Weibull distribution.
    """
    # Define dist with parameters from the distribution fitted to
    # dataset A with the MLE in https://arxiv.org/pdf/1911.12835.pdf .
    dist = ExponentiatedWeibullDistribution(alpha=0.0373, beta=0.4743, delta=46.6078)

    # PDF(0.7) should be roughly 1.1, see Figure 3 in
    # https://arxiv.org/pdf/1911.12835.pdf .
    x = dist.pdf(0.7)
    assert x > 0.6
    assert x < 1.5

    # PDF(2) should be roughly 0.1, see Figure 12
    # in https://arxiv.org/pdf/1911.12835.pdf .
    x = dist.pdf(2)
    assert x > 0.05
    assert x < 0.2

    # PDF(value less than 0) should be 0.
    x = dist.pdf(-1)
    assert x == 0


def test_fitting_exponentiated_weibull():
    """
    Tests estimating the parameters of the  exponentiated Weibull distribution.
    """

    dist = ExponentiatedWeibullDistribution()

    # Draw 1000 samples from a Weibull distribution with shape=1.5 and scale=3,
    # which represents significant wave height.
    hs = sts.weibull_min.rvs(1.5, loc=0, scale=3, size=1000, random_state=42)

    dist.fit(hs, method="wlsq", weights="quadratic")

    # shape parameter/ beta should be about 1.5.
    assert dist.parameters["beta"] > 1
    assert dist.parameters["beta"] < 2

    # scale parameter / alpha should be about 3.
    assert dist.parameters["alpha"] > 2
    assert dist.parameters["alpha"] < 4

    # shape2 parameter / delta should be about 1.
    assert dist.parameters["delta"] > 0.5
    assert dist.parameters["delta"] < 2

    dist = ExponentiatedWeibullDistribution(f_delta=1)

    dist.fit(hs, method="wlsq", weights="quadratic")

    # shape parameter/ beta should be about 1.5.
    assert dist.parameters["beta"] > 1
    assert dist.parameters["beta"] < 2

    # scale parameter / alpha should be about 3.
    assert dist.parameters["alpha"] > 2
    assert dist.parameters["alpha"] < 4

    # shape2 parameter / delta should be 1.
    assert dist.parameters["delta"] == 1

    # Check whether the fitted distribution has a working CDF and PDF.
    assert dist.cdf(2) > 0
    assert dist.pdf(2) > 0


def test_fit_exponentiated_weibull_with_zero():
    """
    Tests fitting the exponentiated Weibull distribution if the dataset
    contains 0s.
    """

    dist = ExponentiatedWeibullDistribution()

    # Draw 1000 samples from a Weibull distribution with shape=1.5 and scale=3,
    # which represents significant wave height.
    hs = sts.weibull_min.rvs(1.5, loc=0, scale=3, size=1000, random_state=42)

    # Add zero-elements to the dataset.
    hs = np.append(hs, [0, 0, 1.3])

    dist.fit(hs, method="wlsq", weights="quadratic")

    assert dist.parameters["beta"] == pytest.approx(1.5, abs=0.5)
    assert dist.parameters["alpha"] == pytest.approx(3, abs=1)
    assert dist.parameters["delta"] == pytest.approx(1, abs=0.5)


def test_generalized_gamma_reproduce_Ochi_CDF():
    """
    Test reproducing the fitting of Ochi (1992) and compare it to
    virocons implementation of the generalized gamma distribution. The results
    should be the same.

    """

    # Define dist with parameters from the distribution of Fig. 4b in
    # M.K. Ochi, New approach for estimating the severest sea state from
    # statistical data , Coast. Eng. Chapter 38 (1992)
    # pp. 512-525.

    dist = GeneralizedGammaDistribution(1.60, 0.98, 1.37)

    # CDF(0.5) should be roughly 0.21, see Fig. 4b
    # CDF(1) should be roughly 0.55, see Fig. 4b
    # CDF(1.5) should be roughly 0.70, see Fig. 4b
    # CDF(2) should be roughly 0.83, see Fig. 4b
    # CDF(4) should be roughly 0.98, see Fig. 4b
    # CDF(6) should be roughly 0.995, see Fig. 4b

    p1 = dist.cdf(0.5)
    p2 = dist.cdf(1)
    p3 = dist.cdf(1.5)
    p4 = dist.cdf(2)
    p5 = dist.cdf(4)
    p6 = dist.cdf(6)

    np.testing.assert_allclose(p1, 0.21, atol=0.05)
    np.testing.assert_allclose(p2, 0.55, atol=0.05)
    np.testing.assert_allclose(p3, 0.70, atol=0.05)
    np.testing.assert_allclose(p4, 0.83, atol=0.05)
    np.testing.assert_allclose(p5, 0.98, atol=0.005)
    np.testing.assert_allclose(p6, 0.995, atol=0.005)

    # CDF(negative value) should be 0
    p = dist.cdf(-1)
    assert p == 0


def test_generalized_gamma_compare_scipy_to_virocon_PDF():
    """
    Test which compares the PDF of a fitted scipy distribution and compare it
    to the PDF of a fitted virocon distribution. The results should be the
    same since virocon uses scipy.

    """

    x = np.linspace(2, 15, num=100)

    # Create sample
    true_m = 5
    true_c = 2.42
    true_loc = 0
    true_lambda_ = 0.5
    gamma_samples = sts.gengamma.rvs(
        a=true_m,
        c=true_c,
        loc=true_loc,
        scale=1 / true_lambda_,
        size=100,
        random_state=42,
    )

    # Fit distribution with virocon
    my_gamma = GeneralizedGammaDistribution()
    my_gamma.fit(gamma_samples, method="mle")
    my_pdf = my_gamma.pdf(x)

    # Fit distribution with scipy
    ref_gamma = sts.gengamma.fit(gamma_samples, floc=0)
    ref_pdf = sts.gengamma.pdf(
        x, ref_gamma[0], ref_gamma[1], ref_gamma[2], ref_gamma[3]
    )

    # Compare results
    for i in range(len(ref_pdf)):
        np.testing.assert_almost_equal(my_pdf[i], ref_pdf[i])


def test_generalized_gamma_compare_scipy_to_virocon_ICDF():
    """
    Test which compares the ICDF of a fitted scipy distribution and compare it
    to the ICDF of a fitted virocon distribution. The results should be the
    same since virocon uses scipy.

    """

    p = np.linspace(0.01, 0.99, num=100)

    # Create sample
    true_m = 5
    true_c = 2.42
    true_loc = 0
    true_lambda_ = 0.5
    gamma_samples = sts.gengamma.rvs(
        a=true_m,
        c=true_c,
        loc=true_loc,
        scale=1 / true_lambda_,
        size=100,
        random_state=42,
    )

    # Fit distribution with virocon
    my_gamma = GeneralizedGammaDistribution()
    my_gamma.fit(gamma_samples, method="mle")
    my_icdf = my_gamma.icdf(p)

    # Fit distribution with scipy
    ref_gamma = sts.gengamma.fit(gamma_samples, floc=0)
    ref_icdf = sts.gengamma.ppf(
        p, ref_gamma[0], ref_gamma[1], ref_gamma[2], ref_gamma[3]
    )

    # Compare results
    for i in range(len(ref_icdf)):
        np.testing.assert_almost_equal(my_icdf[i], ref_icdf[i])


def test_generalized_gamma_compare_scipy_fit_to_virocon_fit():
    """
    Test comparing the fitting of scipy to the fit of virocon. The results
    should be the same since virocon uses scipy.

    """

    # Create sample
    true_m = 5
    true_c = 2.42
    true_loc = 0
    true_lambda_ = 0.5
    gamma_samples = sts.gengamma.rvs(
        a=true_m,
        c=true_c,
        loc=true_loc,
        scale=1 / true_lambda_,
        size=100,
        random_state=42,
    )

    # Fit distribution with virocon
    my_gamma = GeneralizedGammaDistribution()
    my_gamma.fit(gamma_samples, method="mle")

    # Fit distribution with scipy
    ref_gamma = sts.gengamma.fit(gamma_samples, floc=0)

    # Compare results
    assert my_gamma.m == ref_gamma[0]
    assert my_gamma.c == ref_gamma[1]
    assert my_gamma._scale == ref_gamma[3]


@pytest.fixture(
    params=[
        WeibullDistribution,
        LogNormalDistribution,
        NormalDistribution,
        ExponentiatedWeibullDistribution,
        GeneralizedGammaDistribution,
    ]
)
def dist_cls(request):
    return request.param


def test_use_fixed_values_if_provided(dist_cls):
    # create distributions without arguments to get the parameter names
    dist = dist_cls()
    param_names = dist.parameters.keys()
    f_param_names = ["f_" + par_name for par_name in param_names]

    # choose arbitrary values, that should work for all distribution and should be different to the default
    ref_values = [0.42 * (i + 1) for i in range(len(f_param_names))]
    f_params = {
        f_param_name: val for f_param_name, val in zip(f_param_names, ref_values)
    }
    # create distribution with fixed parameters
    dist = dist_cls(**f_params)
    param_values = list(dist.parameters.values())
    # check that the distribution uses the fixed values
    for reference_value, actual_value in zip(ref_values, param_values):
        assert reference_value == pytest.approx(actual_value)


def test_draw_sample(dist_cls):
    dist = dist_cls(1, 1)
    n = 4
    sample = dist.draw_sample(n, random_state=42)
    assert len(sample) == n



"""
Example of how to use the wind speed - wave height model that was proposed in Haselsteiner et al. (2020).

Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben, K.-D. (2020).
Global hierarchical models for wind and wave contours:
Physical interpretations of the dependence functions.
Proc. 39th International Conference on Ocean, Offshore and Arctic Engineering (OMAE 2020).
https://doi.org/10.1115/OMAE2020-18668

This figure is presented as the right panel of Figure 6.4 on page 109 in Haselsteiner,
A. F. (2022). Offshore structures under extreme loads: A methodology to determine
design loads [University of Bremen]. https://doi.org/10.26092/elib/1615
"""

import numpy as np
import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    ExponentiatedWeibullDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    plot_2D_isodensity,
)

# Load wind speed - wave height measurements from the coastDat-2 hindcast.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D.txt")

# Define the marginal distribution for wind speed.
dist_description_v = {
    "distribution": ExponentiatedWeibullDistribution(),
    "intervals": WidthOfIntervalSlicer(2, min_n_points=50),
}


def _logistics4(x, a=1, b=1, c=-1, d=1):
    return a + b / (1 + np.exp(c * (x - d)))


def _alpha3(x, a, b, c, d_of_x):
    return (a + b * x**c) / 2.0445 ** (1 / d_of_x(x))


# Define the conditional distribution for Hs.
logistics_bounds = [(0, None), (0, None), (None, 0), (0, None)]
alpha_bounds = [(0, None), (0, None), (None, None)]
beta_dep = DependenceFunction(
    _logistics4,
    logistics_bounds,
    weights=lambda x, y: y,
)
alpha_dep = DependenceFunction(
    _alpha3,
    alpha_bounds,
    d_of_x=beta_dep,
    weights=lambda x, y: y,
)
dist_description_hs = {
    "distribution": ExponentiatedWeibullDistribution(f_delta=5),
    "conditional_on": 0,
    "parameters": {
        "alpha": alpha_dep,
        "beta": beta_dep,
    },
}

# Create the joint model structure.
dist_descriptions = [dist_description_v, dist_description_hs]
model = GlobalHierarchicalModel(dist_descriptions)

# Define how the model shall be fitted to data and fit it.
fit_description_v = {"method": "wlsq", "weights": "quadratic"}
fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
fit_descriptions = [fit_description_v, fit_description_hs]
model.fit(data, fit_descriptions)

# Print the estimated parameter values.
print(model)

# Analyze the model's goodness of fit with an isodensity plot.
semantics = {
    "names": ["Wind speed", "Significant wave height"],
    "symbols": ["V", "H_s"],
    "units": [
        "m s$^{-1}$",
        "m",
    ],
}
plot_2D_isodensity(model, data, semantics)
# plt.savefig("virocon_omae2020_vhs.pdf", dpi=300, bbox_inches='tight', pad_inches=0)
plt.show()



import pytest
import numpy as np

from virocon import (
    DependenceFunction,
    LogNormalDistribution,
    WeibullDistribution,
    GlobalHierarchicalModel,
)


@pytest.fixture(scope="module")
def vanem2012_seastate_model():
    """
    This joint distribution model described by Vanem and Bitner-Gregersen (2012)
    is widely used in academia. Here, we use it for evaluation.
    DOI: 10.1016/j.apor.2012.05.006
    """

    def _power3(x, a=0.1000, b=1.489, c=0.1901):
        return a + b * x**c

    # A 3-parameter exponential function (a dependence function).
    def _exp3(x, a=0.0400, b=0.1748, c=-0.2243):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]
    power3 = DependenceFunction(_power3, bounds)
    exp3 = DependenceFunction(_exp3, bounds)

    dist_description_0 = {
        "distribution": WeibullDistribution(alpha=2.776, beta=1.471, gamma=0.8888),
    }
    dist_description_1 = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }
    model = GlobalHierarchicalModel([dist_description_0, dist_description_1])

    return model


def test_joint_pdf(vanem2012_seastate_model):
    x = [2.5, 6]  # (hs, tz) in (m, s).
    vanem2012_seastate_model.pdf(x)  # TODO is this needed?

    # Second sea state's period is unphysical due to wave breaking.
    xs = [
        [2.5, 6],
        [2.5, 2],
    ]
    fs = vanem2012_seastate_model.pdf(xs)
    assert fs[0] > fs[1]


def test_joint_cdf(vanem2012_seastate_model):
    # cdf() takes is slow to evaluate (one call ~ 5 s)
    x = [2.5, 6]  # (hs, tz) in (m, s).
    vanem2012_seastate_model.cdf(x)  # TODO is this needed?
    xs = [[2.5, 2], [2.5, 6], [2.5, 10]]
    Fs = vanem2012_seastate_model.cdf(xs)
    assert Fs[1] > Fs[0]
    assert Fs[2] > Fs[1]



import pytest

import pandas as pd
import numpy as np

from pathlib import Path

from virocon import (
    DependenceFunction,
    WeibullDistribution,
    LogNormalDistribution,
    GlobalHierarchicalModel,
    IFORMContour,
)

from virocon.utils import (
    read_ec_benchmark_dataset,
    ROOT_DIR,
    sort_points_to_form_continuous_line,
    calculate_design_conditions,
)


@pytest.fixture(scope="module")
def seastate_model():
    """
    This joint distribution model described by Vanem and Bitner-Gregersen (2012)
    is widely used in academia. Here, we use it for evaluation.
    DOI: 10.1016/j.apor.2012.05.006
    """

    def _power3(x, a=0.1000, b=1.489, c=0.1901):
        return a + b * x**c

    # A 3-parameter exponential function (a dependence function).
    def _exp3(x, a=0.0400, b=0.1748, c=-0.2243):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]
    power3 = DependenceFunction(_power3, bounds)
    exp3 = DependenceFunction(_exp3, bounds)

    dist_description_0 = {
        "distribution": WeibullDistribution(alpha=2.776, beta=1.471, gamma=0.8888),
    }
    dist_description_1 = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }
    model = GlobalHierarchicalModel([dist_description_0, dist_description_1])

    return model


def test_read_ec_benchmark_dataset():
    path = str(ROOT_DIR.joinpath(Path("datasets/ec-benchmark_dataset_A.txt")))
    data = read_ec_benchmark_dataset(path)
    assert isinstance(data, pd.DataFrame)
    assert (
        data.columns == ["significant wave height (m)", "zero-up-crossing period (s)"]
    ).all()
    assert isinstance(data.index, pd.DatetimeIndex)
    assert len(data) == 82805


def test_calculate_design_conditions(seastate_model):
    # Test design condition calculation with  the IFORM contour presented in
    # Haselsteiner et al. (2017; DOI: 10.1016/j.coastaleng.2017.03.002 .
    alpha = 1 / (25 * 365.25 * 24 / 3)
    contour = IFORMContour(seastate_model, alpha)

    design_conditions = calculate_design_conditions(
        contour, steps=None, swap_axis=False
    )
    assert design_conditions.shape == (10, 2)
    np.testing.assert_allclose(design_conditions[1], [2.5, 11.6], atol=0.5)
    np.testing.assert_allclose(design_conditions[9], [15.2, 13.4], atol=0.5)

    steps = 20
    design_conditions = calculate_design_conditions(
        contour, steps=steps, swap_axis=False
    )
    assert design_conditions.shape == (steps, 2)
    np.testing.assert_allclose(design_conditions[1], [1.6, 10.9], atol=0.5)
    np.testing.assert_allclose(design_conditions[19], [15.2, 13.4], atol=0.5)

    design_conditions = calculate_design_conditions(
        contour, steps=steps, swap_axis=True
    )
    assert design_conditions.shape == (steps, 2)
    np.testing.assert_allclose(design_conditions[0], [2.6, 1.1], atol=0.5)
    np.testing.assert_allclose(design_conditions[19], [13.9, 14.1], atol=0.5)


def test_sort_points_to_form_continuous_line():
    phi = np.linspace(0, 1.8 * np.pi, num=10, endpoint=False)
    ref_x = np.cos(phi)
    ref_y = np.sin(phi)

    shuffle_idx = [5, 2, 0, 6, 9, 4, 1, 8, 3, 7]
    rand_x = ref_x[shuffle_idx]
    rand_y = ref_y[shuffle_idx]

    my_x, my_y = sort_points_to_form_continuous_line(
        rand_x, rand_y, search_for_optimal_start=True
    )

    # For some reason this test fails on MacOS (but works on Ubuntu and Windows).
    # In principal, clockwise and anti-clockwise are both correct solutions, but on MacOS
    # it is neither, the starting point is in the middle of the sequency.
    try:
        np.testing.assert_array_equal(my_x, ref_x)
        np.testing.assert_array_equal(my_y, ref_y)
    except AssertionError:
        np.testing.assert_array_equal(my_x[::-1], ref_x)
        np.testing.assert_array_equal(my_y[::-1], ref_y)



"""
Fit two different joint distribution models to a wind speed -
wave height dataset and compare their goodness of fits.
"""

import numpy as np

import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    get_DNVGL_Hs_U,
    get_OMAE2020_V_Hs,
    plot_marginal_quantiles,
    plot_2D_isodensity,
)

# Load dataset (this is dataset D from a benchmarking exercise on environmental
# contours, see https://github.com/ec-benchmark-organizers/ec-benchmark
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D.txt")

# Define the structure of the first joint distribution model. This model
# is recommended in the DNVGL's "Recommended practice DNVGL-RP-C205: Environmental
# conditions and environmental loads." (2017) in section 3.6.4.
dist_descriptions1, fit_descriptions1, semantics1 = get_DNVGL_Hs_U()
model1 = GlobalHierarchicalModel(dist_descriptions1)

# Define the structure of the first joint distribution model. This model
# was proposed at the OMAE 2020 conference by Haselesteiner et al:
# Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben, K.-D. (2020).
# Global hierarchical models for wind and wave contours: Physical interpretations
# of the dependence functions. Proc. 39th International Conference on Ocean,
# Offshore and Arctic Engineering (OMAE 2020). https://doi.org/10.1115/OMAE2020-18668
dist_descriptions2, fit_descriptions2, semantics2 = get_OMAE2020_V_Hs()
model2 = GlobalHierarchicalModel(dist_descriptions2)

# Fit the two models to the data (estimate their parameter values).
# For model 1, we need the variables in the order hs, v (instead of v, hs)
v = data["wind speed (m/s)"].to_numpy()
hs = data["significant wave height (m)"].to_numpy()
hs_v = np.transpose(np.array([hs, v]))
model1.fit(hs_v, fit_descriptions1)
model2.fit(data, fit_descriptions2)

# Analyze the goodness of fit of the marginal distributions with QQ plots.
fig, axs = plt.subplots(1, 2, figsize=[8, 4])
plot_marginal_quantiles(model1, data, semantics=semantics1, axes=axs)
fig.suptitle("DNVGL 2017 model")
fig, axs = plt.subplots(1, 2, figsize=[8, 4])
plot_marginal_quantiles(model2, data, semantics=semantics2, axes=axs)
fig.suptitle("OMAE 2020 model")

# Analyze the overall goodness of fit by plotting isodensity lines.
fig, axs = plt.subplots(1, 2, figsize=[8, 4])
plot_2D_isodensity(model1, hs_v, semantics1, swap_axis=True, ax=axs[0])
plot_2D_isodensity(model2, data, semantics2, ax=axs[1])
axs[0].set_title("DNVGL 2017 model")
axs[1].set_title("OMAE 2020 model")

plt.tight_layout()
plt.show()



import numpy as np
import pandas as pd

from virocon import (
    get_DNVGL_Hs_Tz,
    get_DNVGL_Hs_U,
    get_Windmeier_EW_Hs_S,
    get_Nonzero_EW_Hs_S,
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    TransformedModel,
    IFORMContour,
    variable_transform,
)


def test_DNVGL_Hs_Tz():
    # Load sea state measurements.
    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A_1year.txt")

    # Define the structure of the joint distribution model.
    dist_descriptions, fit_descriptions, semantics = get_DNVGL_Hs_Tz()
    model = GlobalHierarchicalModel(dist_descriptions)

    # Fit the model and compute an IFORM contour.
    model.fit(data, fit_descriptions)
    tr = 20  # Return period in years.
    ts = 1  # Sea state duration in hours.
    alpha = 1 / (tr * 365.25 * 24 / ts)
    contour = IFORMContour(model, alpha)
    coords = contour.coordinates

    # The highest Hs value is ca. 10 m and the highest Tz value is ca. 17 s,
    # see https://github.com/virocon-organization/virocon/pull/228/files#r2007155949
    np.testing.assert_allclose(max(coords[:, 0]), 10, atol=1.5)
    np.testing.assert_allclose(max(coords[:, 1]), 17, atol=2)


def test_DNVGL_Hs_U():
    # Load sea state measurements.
    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D_1year.txt")

    # Switch the order of wind speed and significant wave height in the data frame
    # because the DNVGL model requires the order Hs-U and the original dataset is U-Hs.
    cols = data.columns.tolist()
    cols = cols[-1:] + cols[:1]
    data = data[cols]

    # Define the structure of the joint distribution model.
    dist_descriptions, fit_descriptions, semantics = get_DNVGL_Hs_U()
    model = GlobalHierarchicalModel(dist_descriptions)

    # Fit the model and compute an IFORM contour.
    model.fit(data, fit_descriptions)
    tr = 50  # Return period in years.
    ts = 1  # Sea state duration in hours.
    alpha = 1 / (tr * 365.25 * 24 / ts)
    contour = IFORMContour(model, alpha)
    coords = contour.coordinates

    # Compare with the results from contours presented in Haselsteiner et al. (2019),
    # https://doi.org/10.1115/OMAE2019-96523
    # Highest Hs value is ca. 9.5 m, highest U value is ca. 28 m/s.
    np.testing.assert_allclose(max(coords[:, 0]), 9.5, atol=1)
    np.testing.assert_allclose(max(coords[:, 1]), 28, atol=3)


def test_windmeier_ew_model():
    """
    Test the sea state model proposed by Kai-Lukas Windmeier (DOI: 10.26092/elib/2181).
    This model is defined in Hs-steepness space such that a variable transformation
    to Hs-Tz space is necessary.
    """

    # Load sea state measurements.
    data_hs_tz = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_C_1year.txt")
    hs = data_hs_tz["significant wave height (m)"]
    tz = data_hs_tz["zero-up-crossing period (s)"]
    temp, steepness = variable_transform.hs_tz_to_hs_s(hs, tz)
    steepness.name = "steepness"
    data_hs_s = pd.concat([hs, steepness], axis=1)

    # Define the structure of the joint distribution model.
    (
        dist_descriptions,
        fit_descriptions,
        semantics,
        transformations,
    ) = get_Windmeier_EW_Hs_S()
    model = GlobalHierarchicalModel(dist_descriptions)

    # Fit the model in Hs-S space.
    model.fit(data_hs_s, fit_descriptions)

    f = model.marginal_pdf(1.0, 0)
    assert f > 0.1
    assert f < 2

    p = model.marginal_cdf(1.0, 0)
    assert p > 0.1
    assert p < 0.8

    # Transform the fitted model to Hs-Tz space.
    t_model = TransformedModel(
        model,
        transformations["transform"],
        transformations["inverse"],
        transformations["jacobian"],
        precision_factor=0.2,  # Use low precision to speed up test.
        random_state=42,
    )

    assert "TransformedModel" in str(t_model)
    assert "ExponentiatedWeibullDistribution" in str(t_model)

    p = t_model.cdf([2, 3])
    assert p < 1

    p = t_model.empirical_cdf([2, 3])
    assert p < 1

    try:
        t_model.marginal_pdf(3, 1)
        assert False
    except NotImplementedError:
        assert True

    try:
        t_model.marginal_cdf(3, 1)
        assert False
    except NotImplementedError:
        assert True

    # Compute a contour.
    tr = 1  # Return period in years.
    ts = 1  # Sea state duration in hours.
    alpha = 1 / (tr * 365.25 * 24 / ts)
    contour = IFORMContour(
        t_model, alpha, n_points=20
    )  # Use few points to speed up tests.
    coords = contour.coordinates

    # A test during development of this test.
    # import matplotlib.pyplot as plt
    # from virocon import plot_2D_contour, plot_2D_isodensity
    # plot_2D_contour(contour, data_hs_tz, semantics=semantics, swap_axis=True)
    # plot_2D_isodensity(t_model, data_hs_tz, semantics=semantics, swap_axis=True)
    # plt.show()

    # Reference values are from Windmeier's Master thesis, page 60, DOI: 10.26092/elib/2181
    # Highest Hs values of the contour should be roughly Hs = 7.2 m, Tz = 11 s.
    # Note that in the Master thesis contours for dataset A and B are incorrect.
    np.testing.assert_allclose(max(coords[:, 0]), 7.2, atol=1)
    np.testing.assert_allclose(max(coords[:, 1]), 11, atol=1.5)


def test_nonzero_ew_model():
    """
    Test the nonzero EW sea state model.
    This model is defined in Hs-steepness space such that a variable transformation
    to Hs-Tz space is necessary.
    """

    # Load sea state measurements.
    data_hs_tz = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_C_1year.txt")
    hs = data_hs_tz["significant wave height (m)"]
    tz = data_hs_tz["zero-up-crossing period (s)"]
    temp, steepness = variable_transform.hs_tz_to_hs_s(hs, tz)
    steepness.name = "steepness"
    data_hs_s = pd.concat([hs, steepness], axis=1)

    # Define the structure of the joint distribution model.
    (
        dist_descriptions,
        fit_descriptions,
        semantics,
        transformations,
    ) = get_Nonzero_EW_Hs_S()
    model = GlobalHierarchicalModel(dist_descriptions)

    # Fit the model in Hs-S space.
    model.fit(data_hs_s, fit_descriptions)

    # Transform the fitted model to Hs-Tz space.
    t_model = TransformedModel(
        model,
        transformations["transform"],
        transformations["inverse"],
        transformations["jacobian"],
        precision_factor=0.2,  # Use low precision to speed up test.
        random_state=42,
    )

    # Compute a contour.
    tr = 1  # Return period in years.
    ts = 1  # Sea state duration in hours.
    alpha = 1 / (tr * 365.25 * 24 / ts)
    contour = IFORMContour(
        t_model, alpha, n_points=50
    )  # Use few points to speed up tests.
    coords = contour.coordinates

    # A test during development of this test.
    # import matplotlib.pyplot as plt
    # from virocon import plot_2D_contour, plot_2D_isodensity
    # plot_2D_contour(contour, data_hs_tz, semantics=semantics, swap_axis=True)
    # plot_2D_isodensity(t_model, data_hs_tz, semantics=semantics, swap_axis=True)
    # plt.show()

    np.testing.assert_allclose(max(coords[:, 0]), 7.8, atol=1)
    np.testing.assert_allclose(max(coords[:, 1]), 11.9, atol=1.5)



"""
Example of how to use the Hs-Tz model that was proposed in Haselsteiner et al. (2020).

Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben, K.-D. (2020).
Global hierarchical models for wind and wave contours:
Physical interpretations of the dependence functions.
Proc. 39th International Conference on Ocean, Offshore and Arctic Engineering (OMAE 2020).
https://doi.org/10.1115/OMAE2020-18668

This figure is presented as the left panel of Figure 6.4 on page 109 in Haselsteiner,
A. F. (2022). Offshore structures under extreme loads: A methodology to determine
design loads [University of Bremen]. https://doi.org/10.26092/elib/1615
"""

import numpy as np
import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    ExponentiatedWeibullDistribution,
    LogNormalDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    plot_2D_isodensity,
)

# Load sea state measurements from NDBC buoy 44007.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A.txt")

# Define the marginal distribution for Hs.
dist_description_hs = {
    "distribution": ExponentiatedWeibullDistribution(),
    "intervals": WidthOfIntervalSlicer(width=0.5, min_n_points=50),
}


def _asymdecrease3(x, a, b, c):
    return a + b / (1 + c * x)


def _lnsquare2(x, a, b, c):
    return np.log(a + b * np.sqrt(np.divide(x, 9.81)))


# Define the conditional distribution for Tz.
bounds = [(0, None), (0, None), (None, None)]
sigma_dep = DependenceFunction(_asymdecrease3, bounds=bounds)
mu_dep = DependenceFunction(_lnsquare2, bounds=bounds)
dist_description_tz = {
    "distribution": LogNormalDistribution(),
    "conditional_on": 0,
    "parameters": {
        "sigma": sigma_dep,
        "mu": mu_dep,
    },
}

# Create the joint model structure.
dist_descriptions = [dist_description_hs, dist_description_tz]
model = GlobalHierarchicalModel(dist_descriptions)

# Define how the model shall be fitted to data and fit it.
fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
fit_descriptions = [fit_description_hs, None]
model.fit(data, fit_descriptions)


# Print the estimated parameter values.
print(model)

# Analyze the model's goodness of fit with an isodensity plot.
semantics = {
    "names": ["Significant wave height", "Zero-up-crossing period"],
    "symbols": ["H_s", "T_z"],
    "units": ["m", "s"],
}
plot_2D_isodensity(model, data, semantics, swap_axis=True)
# plt.savefig("virocon_omae2020_hstz.pdf", dpi=300, bbox_inches='tight', pad_inches=0)
plt.show()



"""
Common joint model structure are predefined in this module.
"""

import numpy as np

from virocon import (
    WeibullDistribution,
    LogNormalDistribution,
    ExponentiatedWeibullDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    variable_transform,
)

__all__ = [
    "get_DNVGL_Hs_Tz",
    "get_DNVGL_Hs_U",
    "get_OMAE2020_Hs_Tz",
    "get_OMAE2020_V_Hs",
    "get_Windmeier_EW_Hs_S",
    "get_Nonzero_EW_Hs_S",
]


def get_DNVGL_Hs_Tz():
    """
    Get DNVGL significant wave height and wave period model.

    Get the descriptions necessary to create th significant wave height
    and wave period model as defined in DNVGL [1]_ in section 3.6.3.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : None
        Default fit is used so None is returned.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.

    References
    ----------
    .. [1] DNV GL (2017). Recommended practice DNVGL-RP-C205: Environmental
        conditions and environmental loads.

    """

    # TODO docstrings with links to literature
    # DNVGL 3.6.3
    def _power3(x, a, b, c):
        return a + b * x**c

    def _exp3(x, a, b, c):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]

    power3 = DependenceFunction(_power3, bounds, latex="$a + b * x^c$")
    exp3 = DependenceFunction(_exp3, bounds, latex="$a + b * \exp(c * x)$")

    dist_description_hs = {
        "distribution": WeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(
            width=0.5
        ),  # 0.5 m is a common choice, see e.g. 10.1115/OMAE2020-18041
    }

    dist_description_tz = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }

    dist_descriptions = [dist_description_hs, dist_description_tz]

    fit_descriptions = [{"method": "MoM"}, None]

    semantics = {
        "names": ["Significant wave height", "Zero-up-crossing period"],
        "symbols": ["H_s", "T_z"],
        "units": ["m", "s"],
    }

    return dist_descriptions, fit_descriptions, semantics


def get_DNVGL_Hs_U():
    """
    Get DNVGL significant wave height and wind speed model.

    Get the descriptions necessary to create the significant wave height
    and wind speed model as defined in DNVGL [2]_ in section 3.6.4.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : None
        Default fit is used so None is returned.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.

    References
    ----------
    .. [2] DNV GL (2017). Recommended practice DNVGL-RP-C205: Environmental
        conditions and environmental loads.
    """

    def _power3(x, a, b, c):
        return a + b * x**c

    bounds = [(0, None), (0, None), (None, None)]

    alpha_dep = DependenceFunction(_power3, bounds=bounds, latex="$a + b * x^c$")
    beta_dep = DependenceFunction(_power3, bounds=bounds, latex="$a + b * x^c$")

    dist_description_hs = {
        "distribution": WeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=0.5, min_n_points=20),
    }

    dist_description_u = {
        "distribution": WeibullDistribution(f_gamma=0),
        "conditional_on": 0,
        "parameters": {
            "alpha": alpha_dep,
            "beta": beta_dep,
        },
    }

    dist_descriptions = [dist_description_hs, dist_description_u]

    fit_descriptions = None

    semantics = {
        "names": ["Significant wave height", "Wind speed"],
        "symbols": ["H_s", "U"],
        "units": ["m", "m s$^{-1}$"],
    }

    return dist_descriptions, fit_descriptions, semantics


def get_OMAE2020_Hs_Tz():
    """
    Get OMAE2020 significant wave height and wave period model.

    Get the descriptions necessary to create the significant wave height
    and wave period model as described by Haselsteiner et al. [3]_.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : list of dict
        List of dictionaries containing the fit description for each dimension.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.

    References
    ----------
    .. [3] Haselsteiner, A.F.; Sander, A.; Ohlendorf, J.H.; Thoben, K.D. (2020)
        Global hierarchical models for wind and wave contours: Physical
        interpretations of the dependence functions. OMAE 2020, Fort Lauderdale,
        USA. Proceedings of the 39th International Conference on Ocean,
        Offshore and Arctic Engineering.
    """

    def _asymdecrease3(x, a, b, c):
        return a + b / (1 + c * x)

    def _lnsquare2(x, a, b, c):
        return np.log(a + b * np.sqrt(np.divide(x, 9.81)))

    bounds = [(0, None), (0, None), (None, None)]

    sigma_dep = DependenceFunction(
        _asymdecrease3, bounds=bounds, latex="$a + b / (1 + c * x)$"
    )
    mu_dep = DependenceFunction(
        _lnsquare2, bounds=bounds, latex="$\ln(a + b \sqrt{x / 9.81})$"
    )

    dist_description_hs = {
        "distribution": ExponentiatedWeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=0.5, min_n_points=50),
    }

    dist_description_tz = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {
            "sigma": sigma_dep,
            "mu": mu_dep,
        },
    }

    dist_descriptions = [dist_description_hs, dist_description_tz]

    fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
    fit_descriptions = [fit_description_hs, None]

    semantics = {
        "names": ["Significant wave height", "Zero-up-crossing period"],
        "symbols": ["H_s", "T_z"],
        "units": ["m", "s"],
    }

    return dist_descriptions, fit_descriptions, semantics


def get_OMAE2020_V_Hs():
    """
    Get OMAE2020 wind speed and significant wave height model.

    Get the descriptions necessary to create the wind speed and
    significant wave height model as described by Haselsteiner et al. [4]_.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : list of dict
        List of dictionaries containing the fit description for each dimension.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.

    References
    ----------
    .. [4] Haselsteiner, A.F.; Sander, A.; Ohlendorf, J.H.; Thoben, K.D. (2020)
        Global hierarchical models for wind and wave contours: Physical
        interpretations of the dependence functions. OMAE 2020, Fort Lauderdale,
        USA. Proceedings of the 39th International Conference on Ocean,
        Offshore and Arctic Engineering.
    """

    def _logistics4(x, a=1, b=1, c=-1, d=1):
        return a + b / (1 + np.exp(c * (x - d)))

    def _alpha3(x, a, b, c, d_of_x):
        return (a + b * x**c) / 2.0445 ** (1 / d_of_x(x))

    logistics_bounds = [(0, None), (0, None), (None, 0), (0, None)]

    alpha_bounds = [(0, None), (0, None), (None, None)]

    beta_dep = DependenceFunction(
        _logistics4,
        logistics_bounds,
        weights=lambda x, y: y,
        latex="$a + b / (1 + \exp[c * (x -d)])$",
    )
    alpha_dep = DependenceFunction(
        _alpha3,
        alpha_bounds,
        d_of_x=beta_dep,
        weights=lambda x, y: y,
        latex="$(a + b * x^c) / 2.0445^{1 / F()}$",
    )

    dist_description_v = {
        "distribution": ExponentiatedWeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(2, min_n_points=50),
    }

    dist_description_hs = {
        "distribution": ExponentiatedWeibullDistribution(f_delta=5),
        "conditional_on": 0,
        "parameters": {
            "alpha": alpha_dep,
            "beta": beta_dep,
        },
    }

    dist_descriptions = [dist_description_v, dist_description_hs]

    fit_description_v = {"method": "wlsq", "weights": "quadratic"}
    fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
    fit_descriptions = [fit_description_v, fit_description_hs]

    semantics = {
        "names": ["Wind speed", "Significant wave height"],
        "symbols": ["V", "H_s"],
        "units": [
            "m s$^{-1}$",
            "m",
        ],
    }

    return dist_descriptions, fit_descriptions, semantics


def get_Windmeier_EW_Hs_S():
    """
    Get Windmeier's EW sea state model.

    Get the descriptions necessary to create the significant wave height - steepness
    model that was proposed by Windmeier [5]_. Both, Hs and Steepness follow
    an exponentiated Weibull distribution.

    Because the model is defined in Hs-steepness space it must be transformed to
    Hs-Tz for contour calculation.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : None
        Default fit is used so None is returned.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.
    transformations : dict

    References
    ----------
    .. [5] Windmeier, K.-L. (2022). Modeling the statistical distribution of sea state
    parameters [Master Thesis, University of Bremen]. https://doi.org/10.26092/elib/2181
    """

    def _linear2(x, a=0, b=1):
        return a + b * x

    def _limited_growth2(x, a=0.08, b=1):
        # Some people call this equatoin "negative exponential equation"
        # https://www.statforbiology.com/nonlinearregression/usefulequations#negative_exponential_equation
        return a * (1 - np.exp(-b * x))

    def _transform(hs_tz):
        hs = hs_tz[:, 0]
        tz = hs_tz[:, 1]
        s, _ = variable_transform.hs_tz_to_s_d(hs, tz)
        return np.c_[hs, s]

    def _inv_transform(hs_s):
        hs = hs_s[:, 0]
        s = hs_s[:, 1]
        hs, tz = variable_transform.hs_s_to_hs_tz(hs, s)
        return np.c_[hs, tz]

    def _jacobian(hs_s):
        hs = hs_s[:, 0]
        s = hs_s[:, 1]
        return 2 * variable_transform.factor * hs / s**3

    linear_2_bounds = [(0, None), (0, None)]
    limited_growth2_bounds = [(0, 1), (0, None)]

    linear2 = DependenceFunction(_linear2, bounds=linear_2_bounds)
    limited_growth2 = DependenceFunction(
        _limited_growth2, bounds=limited_growth2_bounds
    )

    dist_description_hs = {
        "distribution": ExponentiatedWeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=0.5, min_n_points=50),
    }

    dist_description_s = {
        "distribution": ExponentiatedWeibullDistribution(f_delta=2.35),
        "conditional_on": 0,
        "parameters": {"alpha": limited_growth2, "beta": linear2},
    }

    dist_descriptions = [dist_description_hs, dist_description_s]

    fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
    fit_descriptions = [fit_description_hs, None]

    transformations = transformations = {
        "transform": _transform,
        "inverse": _inv_transform,
        "jacobian": _jacobian,
    }

    semantics = {
        "names": ["Significant wave height", "Zero-up-crossing period"],
        "symbols": ["H_s", "T_z"],
        "units": ["m", "s"],
    }

    return dist_descriptions, fit_descriptions, semantics, transformations


def get_Nonzero_EW_Hs_S():
    """
    Get the non-zero EW sea state model.

    Get the descriptions necessary to create the significant wave height - steepness
    model that is an adaptation of Windmeier's EW model [5]_. Both, Hs and Steepness follow
    an exponentiated Weibull distribution.

    Compared to Windmeier's EW model, this model has a dependence function for scale
    that evaluates to scale > 0 at hs = 0 m .
    The dependence function reads: shift + a * (1 - np.exp(-b * hs))

    Because the model is defined in Hs-steepness space it must be transformed to
    Hs-Tz for contour calculation.

    This model was introduced in the virocon software, there is research paper
    describing it.

    Returns
    -------
    dist_descriptions : list of dict
        List of dictionaries containing the dist descriptions for each dimension.
        Can be used to create a GlobalHierarchicalModel.
    fit_descriptions : None
        Default fit is used so None is returned.
        Can be passed to fit function of GlobalHierarchicalModel.
    semantics : dict
        Dictionary with a semantic description of the model.
        Can be passed to plot functions.
    transformations : dict

    References
    ----------
    .. [5] Windmeier, K.-L. (2022). Modeling the statistical distribution of sea state
    parameters [Master Thesis, University of Bremen]. https://doi.org/10.26092/elib/2181
    """

    def _linear2(x, a=0, b=1):
        return a + b * x

    def _limited_growth_with_shift2(x, a=0.08, b=1):
        # Compared to Windmeier's EW model the idea here is to have f(0) > 0.
        # The idea is based on on the Figure 4.10 in Windmeier's thesis.
        # DOI: 10.26092/elib/2181 .
        # The value of 0.006 is chosen by extrapolating the median scale value line in
        # Figure 4.10 to hs = 0.
        shift = 0.006

        # Some people call this equation "negative exponential equation" when the
        # shift is zero, https://www.statforbiology.com/nonlinearregression/usefulequations#negative_exponential_equation
        return shift + a * (1 - np.exp(-b * x))

    def _transform(hs_tz):
        hs = hs_tz[:, 0]
        tz = hs_tz[:, 1]
        s, _ = variable_transform.hs_tz_to_s_d(hs, tz)
        return np.c_[hs, s]

    def _inv_transform(hs_s):
        hs = hs_s[:, 0]
        s = hs_s[:, 1]
        hs, tz = variable_transform.hs_s_to_hs_tz(hs, s)
        return np.c_[hs, tz]

    def _jacobian(hs_s):
        hs = hs_s[:, 0]
        s = hs_s[:, 1]
        return 2 * variable_transform.factor * hs / s**3

    linear_2_bounds = [(0, None), (0, None)]
    limited_growth_with_shift2_bounds = [(0, 1), (0, None)]

    linear2 = DependenceFunction(_linear2, bounds=linear_2_bounds)
    limited_growth_with_shift2 = DependenceFunction(
        _limited_growth_with_shift2, bounds=limited_growth_with_shift2_bounds
    )

    dist_description_hs = {
        "distribution": ExponentiatedWeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=0.5, min_n_points=50),
    }

    dist_description_s = {
        "distribution": ExponentiatedWeibullDistribution(f_delta=2.35),
        "conditional_on": 0,
        "parameters": {"alpha": limited_growth_with_shift2, "beta": linear2},
    }

    dist_descriptions = [dist_description_hs, dist_description_s]

    fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
    fit_descriptions = [fit_description_hs, None]

    transformations = transformations = {
        "transform": _transform,
        "inverse": _inv_transform,
        "jacobian": _jacobian,
    }

    semantics = {
        "names": ["Significant wave height", "Zero-up-crossing period"],
        "symbols": ["H_s", "T_z"],
        "units": ["m", "s"],
    }

    return dist_descriptions, fit_descriptions, semantics, transformations



"""
Various functions that do not fit in an existing module.
"""

import networkx as nx
import pandas as pd
import numpy as np

from pathlib import Path

from sklearn.neighbors import NearestNeighbors

from virocon._intersection import intersection
from typing import Union
import numpy.typing as npt

__all__ = [
    "read_ec_benchmark_dataset",
    "calculate_design_conditions",
    "sort_points_to_form_continuous_line",
]

ROOT_DIR = Path(__file__).parent.parent


def read_ec_benchmark_dataset(file_path=None):
    """
    Reads an EC benchmark dataset.

    Reads a text/csv file as it is used in the EC benchmark:
    https://github.com/ec-benchmark-organizers/ec-benchmark

    Parameters
    ----------
    file_path : str
        Path to the dataset file. Defaults to the example dataset A.

    Returns
    -------
    dataset : pandas.DataFrame
        The dataset stored in a pandas DataFrame.
        Use dataset.values to access the underlying numpy array.
    """
    if file_path is None:
        file_path = str(ROOT_DIR.joinpath(Path("datasets/ec-benchmark_dataset_A.txt")))

    data = pd.read_csv(file_path, sep=";", skipinitialspace=True)
    data.index = pd.to_datetime(data.pop(data.columns[0]), format="%Y-%m-%d-%H")
    return data


def calculate_design_conditions(
    contour, steps: Union[list, int] = None, swap_axis=False
) -> npt.ArrayLike:
    """Calculates design conditions (relevant points for structural analysis) for a contour.

    For example, in a sea state contour, one is interested in the upper Hs part for
    a given Tz/Tp. And one often likes to have points that are evenly spaced in
    the Tz/Tp direction, e.g. one Hs-Tz design condition per 0.5 s of Tz.

    Args:
        contour (Contour): Environmental conntour object.
        steps (list of floats or int, optional): Either a list of design conditoin x-values or the number how many
            steps shall be used (then the x-values will be evenly spaced). Defaults to None.
        swap_axis (bool, optional): Wheter x and y shall be swapped. Defaults to False.

    Returns:
        npt.ArrayLike: design conditions (points for structural analysis calculations).
    """
    if swap_axis:
        x_idx = 1
        y_idx = 0
    else:
        x_idx = 0
        y_idx = 1

    coords = contour.coordinates

    x1 = np.append(coords[:, x_idx], coords[0, x_idx])
    y1 = np.append(coords[:, y_idx], coords[0, y_idx])

    # Define the x-positions where a design condition will be calculated
    # The y-values will be found by calculating the intersections between
    # vertical lines and the contour coordinates.
    small_spacer = 0.0001 * (
        np.max(x1) - np.min(x1)
    )  # Required to ensure that conditions on limits are picked up.
    default_lower_limit = np.min(x1) + small_spacer
    default_uppper_limit = np.max(x1) - small_spacer
    if steps is None:
        steps = np.linspace(
            default_lower_limit, default_uppper_limit, endpoint=True, num=10
        )
    else:
        try:
            iter(steps)  # if steps is iterable use it
        except TypeError:  # if steps is not iterable assume it's an int
            steps = np.linspace(
                default_lower_limit, default_uppper_limit, endpoint=True, num=steps
            )

    # Vertical line ylimits.
    y2 = [np.min(y1) - np.max(y1) * 0.1, np.max(y1) + np.max(y1) * 0.1]

    frontier_x = []
    frontier_y = []

    for x2 in steps:
        x, y = intersection(x1, y1, [x2, x2], y2)
        assert len(x) <= 2
        assert len(y) <= 2

        if len(y) == 0:
            continue

        frontier_x.append(x2)
        frontier_y.append(np.max(y))

    design_conditions = np.c_[frontier_x, frontier_y]
    return design_conditions


def sort_points_to_form_continuous_line(x, y, search_for_optimal_start=False):
    """
    Sorts contour points to form a a continous line / contour.

    This function simply sorts 2 dimensional points. The points are given
    by x and y coordinates. This function is used to sort the coordinates
    of 2 dimensional contours.

    Thanks to https://stackoverflow.com/a/37744549

    Parameters
    ----------
    x : array_like
        x-coordinate of contour points.
    y : array_like
        y-coordinate of contour points.
    search_for_optimal_start : boolean, optional
        If True, the algorithm also searches for the ideal starting node, see the
        stackoverflow link for more info. Defaults to False.

    Returns
    -------
    sorted_points : tuple of array_like floats
        The sorted points.
    """

    points = np.c_[x, y]
    clf = NearestNeighbors(n_neighbors=2).fit(points)
    G = clf.kneighbors_graph()
    T = nx.from_scipy_sparse_array(G)
    order = list(nx.dfs_preorder_nodes(T, 0))

    xx = x[order]
    yy = y[order]

    if search_for_optimal_start:
        paths = [list(nx.dfs_preorder_nodes(T, i)) for i in range(len(points))]
        mindist = np.inf
        minidx = 0

        for i in range(len(points)):
            p = paths[i]  # Order of nodes.
            ordered = points[p]  # Ordered nodes.
            # Find cost of that order by the sum of euclidean distances
            # between points (i) and (i + 1).
            cost = (((ordered[:-1] - ordered[1:]) ** 2).sum(1)).sum()
            if cost < mindist:
                mindist = cost
                minidx = i

        opt_order = paths[minidx]

        xx = x[opt_order]
        yy = y[opt_order]

    return xx, yy



import pytest
import numpy as np
import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    IFORMContour,
    DependenceFunction,
    LogNormalDistribution,
    WeibullDistribution,
    GlobalHierarchicalModel,
    get_OMAE2020_V_Hs,
    plot_dependence_functions,
    plot_2D_isodensity,
    plot_2D_contour,
    plot_histograms_of_interval_distributions,
)


@pytest.fixture(scope="module")
def seastate_model():
    """
    This joint distribution model described by Vanem and Bitner-Gregersen (2012)
    is widely used in academia. Here, we use it for evaluation.
    DOI: 10.1016/j.apor.2012.05.006
    """

    def _power3(x, a=0.1000, b=1.489, c=0.1901):
        return a + b * x**c

    def _exp3(x, a=0.0400, b=0.1748, c=-0.2243):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]
    power3 = DependenceFunction(_power3, bounds, latex="$a + b * x^{c}$")
    exp3 = DependenceFunction(_exp3, bounds)

    dist_description_0 = {
        "distribution": WeibullDistribution(alpha=2.776, beta=1.471, gamma=0.8888),
    }
    dist_description_1 = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }
    model = GlobalHierarchicalModel([dist_description_0, dist_description_1])

    return model


@pytest.fixture(scope="module")
def data_used_for_fitting_model():
    return read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D_1year.txt")


@pytest.fixture(scope="module")
def fitted_model(data_used_for_fitting_model):
    """
    Here we fit the joint distribution model described by Haselsteiner et al. (2020)
    to a dataset. We will use this model for various plot tests.
    """
    dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()
    model = GlobalHierarchicalModel(dist_descriptions)

    model.fit(data_used_for_fitting_model, fit_descriptions)

    return model


@pytest.fixture(scope="module")
def semantics_fitted_model():
    """
    Semantics dictionary for the fitted_model.
    """
    dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()

    return semantics


def test_plot_dependence_function(seastate_model, fitted_model, semantics_fitted_model):
    plot_dependence_functions(seastate_model)
    semantics = {
        "names": ["Significant wave height", "Zero-up-crossing wave period"],
        "symbols": ["H_s", "T_z"],
        "units": ["m", "s"],
    }
    plot_dependence_functions(seastate_model, semantics)
    plot_dependence_functions(
        fitted_model,
        semantics_fitted_model,
        par_rename={"alpha": r"$\alpha$", "beta": r"$\beta$"},
    )
    # plt.show()
    plt.close("all")


def test_plot_2D_isodensity():
    dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()
    model = GlobalHierarchicalModel(dist_descriptions)
    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D_1year.txt")
    model.fit(data, fit_descriptions)
    plot_2D_isodensity(model, data)
    limits = [(0, 40), (0, 15)]
    levels = [0.00001, 0.0001, 0.001]
    plot_2D_isodensity(model, data, limits=limits, levels=levels)
    # plt.show()
    plt.close("all")


def test_plot_2D_contour():
    dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()
    model = GlobalHierarchicalModel(dist_descriptions)
    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D_1year.txt")
    model.fit(data, fit_descriptions)
    contour = IFORMContour(model, alpha=0.001)
    plot_2D_contour(contour)
    # plt.show()
    plt.close("all")


def test_plot_histograms_of_interval_distributions(
    data_used_for_fitting_model, fitted_model, semantics_fitted_model
):
    plot_histograms_of_interval_distributions(
        fitted_model, data_used_for_fitting_model, semantics_fitted_model
    )
    # plt.show()
    plt.close("all")



"""
Variable transformation for Hs, Tz, and steepness.

These transformations are used in Windmeier's EW and in the
nonzero EW model (see doi.org/10.26092/elib/2181 and predefined.py)
"""

import numpy as np

g = 9.81
factor = 2 * np.pi / g
factor_sqrt = np.sqrt(factor)


def hs_tz_to_s_d(hs, tz):
    global factor
    s = factor * hs / (tz * tz)
    d = np.sqrt(hs * hs + tz * tz / 2)
    return s, d


def s_d_to_hs_tz(s, d):
    global factor
    global factor_sqrt
    hs = (np.sqrt(16 * d**2 * s**2 + factor**2) - factor) / (4 * s)
    tz = (
        1
        / 2
        * np.sqrt(
            (factor * np.sqrt(16 * d**2 * s**2 + factor**2)) / s**2 - factor**2 / s**2
        )
    )
    return hs, tz


def hs_tz_to_hs_s(hs, tz):
    global factor
    s = factor * hs / (tz * tz)
    return hs, s


def hs_s_to_hs_tz(hs, s):
    global factor_sqrt
    tz = factor_sqrt * np.sqrt(hs / s)
    return hs, tz


def hs_tz_to_s_tz(hs, tz):
    global factor
    s = factor * hs / (tz * tz)
    return s, tz


def s_tz_to_hs_tz(s, tz):
    global factor
    hs = s * np.square(tz) / factor
    return hs, tz



import numpy as np

from virocon import (
    hs_s_to_hs_tz,
    hs_tz_to_hs_s,
    hs_tz_to_s_d,
    hs_tz_to_s_tz,
    s_tz_to_hs_tz,
    s_d_to_hs_tz,
)


def test_variable_transformations():
    n = 100
    hs = np.linspace(0.5, 10, num=n)
    tz = np.linspace(2, 15, num=n)

    hs_hat, tz_hat = s_d_to_hs_tz(*hs_tz_to_s_d(hs, tz))
    np.testing.assert_allclose(hs_hat, hs)
    np.testing.assert_allclose(tz_hat, tz)

    hs_hat, tz_hat = hs_tz_to_hs_s(*hs_s_to_hs_tz(hs, tz))
    np.testing.assert_allclose(hs_hat, hs)
    np.testing.assert_allclose(tz_hat, tz)

    hs_hat, tz_hat = hs_tz_to_s_tz(*s_tz_to_hs_tz(hs, tz))
    np.testing.assert_allclose(hs_hat, hs)
    np.testing.assert_allclose(tz_hat, tz)



"""
Computes IFORM sea state contours using all predefined Hs-Tz joint model structures.
"""

import matplotlib.pyplot as plt
import pandas as pd
import copy

from virocon import (
    read_ec_benchmark_dataset,
    variable_transform,
    get_Windmeier_EW_Hs_S,
    get_Nonzero_EW_Hs_S,
    get_OMAE2020_Hs_Tz,
    get_DNVGL_Hs_Tz,
    GlobalHierarchicalModel,
    TransformedModel,
    IFORMContour,
    plot_2D_contour,
)

# Load sea state measurements.
fname = "datasets/ec-benchmark_dataset_A_1year.txt"
data_hs_tz = read_ec_benchmark_dataset(fname)
hs = data_hs_tz["significant wave height (m)"]
tz = data_hs_tz["zero-up-crossing period (s)"]
temp, steepness = variable_transform.hs_tz_to_hs_s(hs, tz)
steepness.name = "steepness"
data_hs_s = pd.concat([hs, steepness], axis=1)

# Properties for contour calculation.
tr = 1  # Return period in years.
ts = 1  # Sea state duration in hours.
alpha = 1 / (tr * 365.25 * 24 / ts)
n_contour_points = 50

# Properties for Monte Carlo methods.
precision_factor = 0.2  # Use low precision to speed up this example.
random_state = 42

# Compute a contour based on on a DNV model.
print("Calculating contour 1/4 (with DNVGL model)")
(
    dnv_dist_descriptions,
    dnv_fit_descriptions,
    dnv_semantics,
) = get_DNVGL_Hs_Tz()
dnv_model = GlobalHierarchicalModel(dnv_dist_descriptions)
dnv_model.fit(data_hs_tz, fit_descriptions=dnv_fit_descriptions)
dnv_contour = IFORMContour(dnv_model, alpha)


# Compute a contour based on a OMAE2020 model.
print("Calculating contour 2/4 (with OMAE2020 model)")
(
    omae2020_dist_descriptions,
    omae2020_fit_descriptions,
    omae2020_semantics,
) = get_OMAE2020_Hs_Tz()
omae2020_model = GlobalHierarchicalModel(omae2020_dist_descriptions)
omae2020_model.fit(data_hs_tz, fit_descriptions=omae2020_fit_descriptions)
omae2020_contour = IFORMContour(omae2020_model, alpha)

# Compute a contour based on Windmeier's originial EW model.
# Before contour calculation the model is transformed from Hs-S variable space to Hs-Tz.
# Contour calculation using this transfromed joint distribution is numerically expensive.
# Alternatively, we could calculate the contour in Hs-S variable space and then transfer
# the contour coordinates to Hs-Tz. This would be computationally faster and this approach
# is shown with the next model ('non-zero EW model').
print(
    "Calculating contour 3/4 (with Windmeier's EW model). "
    "Calculating this contour will take about 5 minutes because "
    "a Monte Carlo method is used for evaluating the ICDF of the TransFormedModel."
)
(
    dist_descriptions,
    fit_descriptions,
    hs_tz_semantics,
    transformations,
) = get_Windmeier_EW_Hs_S()
windmeier_ew_model = GlobalHierarchicalModel(dist_descriptions)
windmeier_ew_model.fit(data_hs_s, fit_descriptions)  # Fit the model in Hs-S space.
hs_s_semantics = {
    "names": ["Significant wave height", "Steepness"],
    "symbols": ["H_s", "S"],
    "units": ["m", "-"],
}
windmeier_t_model = TransformedModel(  # Transform the fitted model to Hs-Tz space.
    windmeier_ew_model,
    transformations["transform"],
    transformations["inverse"],
    transformations["jacobian"],
    precision_factor=precision_factor,
    random_state=random_state,
)
windmeier_model_contour = IFORMContour(
    windmeier_t_model, alpha, n_points=n_contour_points
)

# Compute a contour based on the non-zero EW model.
# The contour is calculated in Hs-S variable space and then transformed to Hs-Tz.
print(
    "Calculating contour 4/4 (with non-zero EW model). "
    "Calculating this contour will be fast because we compute the contour in the "
    "original Hs-S variable space and then transform the coordinates to Hs-Tz space."
)
(
    dist_descriptions,
    fit_descriptions,
    hs_tz_semantics,
    transformations,
) = get_Nonzero_EW_Hs_S()
ew_model = GlobalHierarchicalModel(dist_descriptions)
ew_model.fit(data_hs_s, fit_descriptions)
ew_t_model = TransformedModel(
    ew_model,
    transformations["transform"],
    transformations["inverse"],
    transformations["jacobian"],
    precision_factor=precision_factor,
    random_state=random_state,
)
ew_model_contour_hs_s = IFORMContour(ew_model, alpha, n_points=n_contour_points)
ew_model_contour_hs_tz = copy.deepcopy(ew_model_contour_hs_s)
ew_model_contour_hs_tz.coordinates = transformations["inverse"](
    ew_model_contour_hs_tz.coordinates
)


# Plot the contours on top of the metocean data.
fig, ax = plt.subplots(1, 1, figsize=[5, 5])
plot_2D_contour(
    dnv_contour, data_hs_tz, semantics=hs_tz_semantics, swap_axis=True, ax=ax
)
plot_2D_contour(
    omae2020_contour, data_hs_tz, semantics=hs_tz_semantics, swap_axis=True, ax=ax
)
plot_2D_contour(
    windmeier_model_contour,
    data_hs_tz,
    semantics=hs_tz_semantics,
    swap_axis=True,
    ax=ax,
)
plot_2D_contour(
    ew_model_contour_hs_tz, data_hs_tz, semantics=hs_tz_semantics, swap_axis=True, ax=ax
)
ax.lines[1].set_color("orchid")
ax.lines[2].set_color("blue")
ax.lines[3].set_color("gray")
ax.legend(
    [ax.lines[0], ax.lines[1], ax.lines[2], ax.lines[3]],
    ["DNV model", "OMAE2020 model", "Windmeier's EW model", "Nonzero EW model"],
    ncol=1,
    frameon=False,
)
ax.set_title(f"{tr}-year IFORM contour")

plt.show()



"""
Model fitting (parameter estimation).
"""

import numpy as np

from scipy.optimize import minimize, curve_fit


def fit_function(func, x, y, p0, method, bounds, weights=None):
    if bounds is not None:
        bounds = convert_bounds_for_curve_fit(bounds)

    if method == "lsq":
        if bounds is None:
            popt, _ = curve_fit(func, x, y, p0)
        else:
            popt, _ = curve_fit(func, x, y, p0, bounds=bounds)
    elif method == "wlsq":
        if bounds is None:
            popt, _ = curve_fit(func, x, y, p0, sigma=weights)
        else:
            popt, _ = curve_fit(func, x, y, p0, sigma=weights, bounds=bounds)
    else:
        raise ValueError(
            "method must be either lsq for least squares or"
            "wlsq for weighted least squares"
        )
    return popt


def convert_bounds_for_curve_fit(bounds):
    lower_bounds = []
    upper_bounds = []
    for lower, upper in bounds:
        lower_bounds.append(lower if lower is not None else -np.inf)
        upper_bounds.append(upper if upper is not None else np.inf)
    return [lower_bounds, upper_bounds]


def get_least_squares_error_func(func, x, y):
    def least_squares_error_func(p):
        return np.sum((func(x, *p) - y) ** 2)

    return least_squares_error_func


def bounds_to_constraints(bounds):
    #  https://stackoverflow.com/a/41761740
    cons = []
    for factor in range(len(bounds)):
        lower, upper = bounds[factor]
        lower = -np.inf if lower is None else lower
        upper = np.inf if upper is None else upper
        lo = {"type": "ineq", "fun": lambda x, lb=lower, i=factor: x[i] - lb}
        up = {"type": "ineq", "fun": lambda x, ub=upper, i=factor: ub - x[i]}
        cons.append(lo)
        cons.append(up)
    return cons


def fit_constrained_function(func, x, y, p0, method, bounds, constraints, weights=None):
    if method == "lsq":
        error_func = get_least_squares_error_func(func, x, y)
    else:
        # TODO implement WLSQ
        raise NotImplementedError(
            "At this time only least squares (lsq) fitting is supported."
        )

    if constraints is None:
        constraints = []

    result = minimize(
        error_func,
        p0,
        method="SLSQP",
        bounds=bounds,
        options={"eps": 1e-15},
    )
    if not result.success:
        raise RuntimeError(
            "Error during fitting in scipy.optimize.minimize. "
            f"Error message was: \n {result.message}."
        )
    popt = result.x
    return popt



"""
Brief example that computes a sea state contour.
"""

import matplotlib.pyplot as plt
from virocon import (
    read_ec_benchmark_dataset,
    get_OMAE2020_Hs_Tz,
    GlobalHierarchicalModel,
    IFORMContour,
    plot_2D_contour,
)

# Load sea state measurements.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A.txt")

# Define the structure of the joint distribution model.
dist_descriptions, fit_descriptions, semantics = get_OMAE2020_Hs_Tz()
model = GlobalHierarchicalModel(dist_descriptions)

# Estimate the model's parameter values (fitting).
model.fit(data, fit_descriptions=fit_descriptions)

# Compute an IFORM contour with a return period of 50 years.
tr = 50  # Return period in years.
ts = 1  # Sea state duration in hours.
alpha = 1 / (tr * 365.25 * 24 / ts)
contour = IFORMContour(model, alpha)

# Plot the contour.
plot_2D_contour(contour, data, semantics=semantics, swap_axis=True)
plt.show()



"""
Functions to plot distributions and contours.
"""

import re
from functools import partial

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as sts

from matplotlib.colors import LinearSegmentedColormap

from virocon.utils import calculate_design_conditions

__all__ = [
    "plot_marginal_quantiles",
    "plot_dependence_functions",
    "plot_histograms_of_interval_distributions",
    "plot_2D_isodensity",
    "plot_2D_contour",
]


# colors (schemes) chosen according to https://personal.sron.nl/~pault/


def _rainbow_PuRd():
    """
    Thanks to Paul Tol (https://personal.sron.nl/~pault/data/tol_colors.py)
    License:  Standard 3-clause BSD
    """
    clrs = [
        "#6F4C9B",
        "#6059A9",
        "#5568B8",
        "#4E79C5",
        "#4D8AC6",
        "#4E96BC",
        "#549EB3",
        "#59A5A9",
        "#60AB9E",
        "#69B190",
        "#77B77D",
        "#8CBC68",
        "#A6BE54",
        "#BEBC48",
        "#D1B541",
        "#DDAA3C",
        "#E49C39",
        "#E78C35",
        "#E67932",
        "#E4632D",
        "#DF4828",
        "#DA2222",
    ]
    cmap = LinearSegmentedColormap.from_list("rainbow_PuRd", clrs)
    cmap.set_bad("#FFFFFF")
    return cmap


# TODO move to utility as it is also used in contours.py
def get_default_semantics(n_dim):
    """
    Generate default semantics for n_dim dimensions.

    Parameters
    ----------
    n_dim : int
        Number of dimensions. Indicating the number of variables of the model.

    Returns
    -------
    semantics: dict
        Generated model description.

    """

    semantics = {
        "names": [f"Variable {dim + 1}" for dim in range(n_dim)],
        "symbols": [f"X_{dim + 1}" for dim in range(n_dim)],
        "units": ["arb. unit" for dim in range(n_dim)],
    }
    return semantics


def _get_n_axes(n_intervals, max_cols=4):
    if n_intervals <= 0:
        raise ValueError(
            f"n_intervals should be a positive integer, but got {n_intervals}."
        )
    elif n_intervals < max_cols**2:
        table = np.array(
            [
                (rows, cols)
                for cols in range(1, max_cols + 1)
                for rows in range(1, cols + 1)
            ]
        )
        table = table[np.prod(table, axis=1) >= n_intervals]
        (nrows, ncols) = table[np.argmin(np.prod(table, axis=1) - n_intervals)]
    else:
        nrows = np.ceil(n_intervals / max_cols).astype(int)
        ncols = max_cols

    fig, axes = plt.subplots(
        nrows=nrows, ncols=ncols, sharex=True, sharey=True, squeeze=False
    )
    return fig, axes.ravel()


def plot_marginal_quantiles(model, sample, semantics=None, axes=None):
    """
    Plot all marginal quantiles of a model.

    Plots the fitted marginal distribution versus a dataset in a quantile-quantile (QQ) plot.

    Parameters
    ----------
    model :  MultivariateModel
        The model used to plot the marginal quantiles.
    sample : ndarray of floats
        The environmental data sample that should be plotted against the fit.
        Shape: (sample_size, n_dim)
    semantics: dict, optional
        The description of the model. If None (the default) generic semantics
        will be used. The structure is as follows:
        modeldesc = {
        "names" : [<Names of variables>],
        "symbols" : [<Description of symbols>],
        "units" : [<Units of variables>] }
    axes: list, optional
        The matplotlib axes objects to plot into. One for each dimension. If
        None (the default) a new figure will be created for each dimension.

    Returns
    -------
    The used matplotlib axes object.

    Notes
    -----
    When saving the resulting axes in a vector image format (pdf, svg) the
    `sample` will still be rasterized to reduce the file size.
    To prevent that, iterate over the axes and use
    ``ax.get_lines()[0].set_rasterized(False)``.

    """

    sample = np.asarray(sample)
    n_dim = model.n_dim
    if semantics is None:
        semantics = get_default_semantics(n_dim)

    if axes is None:
        axes = []
        for i in range(n_dim):
            _, ax = plt.subplots()
            axes.append(ax)

    # probplot expects an object that has a ppf method, but we name it icdf
    # therefor we create a wrapper that maps the ppf to the icdf method
    class MarginalDistWrapper:
        def __init__(self, model, idx):
            self.model = model
            self.idx = idx

        def ppf(self, q):
            return self.model.marginal_icdf(q, self.idx)

    for dim in range(n_dim):
        dist_wrapper = MarginalDistWrapper(model, dim)
        ax = axes[dim]

        sts.probplot(sample[:, dim], dist=dist_wrapper, fit=False, plot=ax)
        ax.get_lines()[0].set_markerfacecolor("k")
        ax.get_lines()[0].set_markeredgecolor("k")
        ax.get_lines()[0].set_marker("x")
        ax.get_lines()[0].set_markersize(3)

        # Because a sample usually holds much more than 1000 observations, the
        # output shall be rasterized to reduce file size if the figure is saved
        # in a vector file format (svg, pdf).
        ax.get_lines()[0].set_rasterized(True)

        # prior to scipy version 1.7.0 a regression fit line was plotted,
        # even with option fit=False. As discussed in PR#149 we do not want
        # to keep this line. So here we remove it if it was plotted.
        if len(ax.lines) > 1:
            ax.lines[1].remove()

        # draw 45° line
        # adapted from statsmodels.graphics.gofplots.qqline
        # https://github.com/statsmodels/statsmodels/blob/161ca84b2b6e2b3ba5ef6b570d8c907c9b06f5de/statsmodels/graphics/gofplots.py#L892-L897
        end_pts = list(zip(ax.get_xlim(), ax.get_ylim()))
        end_pts[0] = min(end_pts[0])
        end_pts[1] = max(end_pts[1])
        ax.plot(end_pts, end_pts, c="#BB5566")
        ax.set_xlim(end_pts)
        ax.set_ylim(end_pts)

        name_and_unit = f"{semantics['names'][dim].lower()} ({semantics['units'][dim]})"
        ax.set_xlabel(f"Theoretical quantiles of {name_and_unit}")
        ax.set_ylabel(f"Ordered values of {name_and_unit}")
        ax.title.set_text("")

    return axes


def plot_dependence_functions(model, semantics=None, par_rename={}, axes=None):
    """
    Plot the fitted dependence functions of a model.

    Parameters
    ----------
    model :  MultivariateModel
        The model with the fitted dependence functions.
    semantics: dict, optional
        The description of the model. If None (the default) generic semantics will be used.
    par_rename : dict, optional
        A dictionary that maps from names of conditional parameters to a
        string. If e.g. the model has a distribution with a conditional
        parameter named 'mu' one could change that in the plot to '$mu$' with
        {'mu': '$mu$'}.
    axes: list of matplotlib axes objects, optional
        If not further specified, the number of axes are dependent on the number
        of dimensions of the model. Defaults to None.

    Returns
    -------
    The used matplotlib axes object.

    """

    n_dim = model.n_dim
    conditional_dist_idc = [
        dim for dim in range(n_dim) if model.conditional_on[dim] is not None
    ]

    if semantics is None:
        semantics = get_default_semantics(n_dim)

    if axes is None:
        n_axes = 0
        for dim in conditional_dist_idc:
            n_axes += len(model.distributions[dim].conditional_parameters)
        axes = []
        for i in range(n_axes):
            _, ax = plt.subplots()
            axes.append(ax)

    axes_counter = 0
    for dim in conditional_dist_idc:
        dist = model.distributions[dim]
        conditioning_values = dist.conditioning_values
        if conditioning_values is not None:
            x = np.linspace(0, max(conditioning_values))
        else:
            x = np.linspace(0, 10)
        cond_idx = model.conditional_on[dim]
        x_name = semantics["names"][cond_idx]
        x_symbol = semantics["symbols"][cond_idx]
        x_unit = semantics["units"][cond_idx]
        x_label = f"{x_name}," + r" $\it{" + f"{x_symbol}" + r"}$" + f" ({x_unit})"
        for par_name, dep_func in dist.conditional_parameters.items():
            ax = axes[axes_counter]
            axes_counter += 1
            # If a model is created directly (without fitting it to a dataset), it does
            # not have conditioning values.
            if conditioning_values is not None:
                par_values = [par[par_name] for par in dist.parameters_per_interval]
                ax.scatter(
                    conditioning_values,
                    par_values,
                    c="k",
                    marker="x",
                    label="estimates from intervals",
                )
            if dep_func.latex is not None:
                dep_func_label = dep_func.latex

                # Replace 'x' with variable symbol (e.g. 'h_s')
                splitted_symbol = x_symbol.split("_")
                if len(splitted_symbol) == 1:  # If there was no underscore.
                    var_symbol = splitted_symbol[0].lower()
                else:  # If there was one or many underscores.
                    var_symbol = (
                        splitted_symbol[0].lower() + "_" + "_".join(splitted_symbol[1:])
                    )

                # Replace 'x' if it is not part of '\exp' which is checked by checking whether
                # 'x' follows '\e'.
                dep_func_label = re.sub(
                    r"(?<!\\e)x", "{" + var_symbol + "}", dep_func_label
                )

                # Replace symbols of parameters (a, b, ..) with estimated values.
                for par_name_local, par_value_local in dep_func.parameters.items():
                    dep_func_label = dep_func_label.replace(
                        par_name_local, "{" + "{:.2f}".format(par_value_local) + "}"
                    )
            else:
                if not isinstance(dep_func.func, partial):
                    dep_func_label = "Dependence function: " + dep_func.func.__name__
                else:
                    dep_func_label = (
                        "Dependence function: " + dep_func.func.func.__name__
                    )
            ax.plot(x, dep_func(x), c="#004488", label=dep_func_label)
            ax.set_xlabel(x_label)
            if par_name in par_rename:
                y_label = par_rename[par_name]
            else:
                y_label = par_name
            ax.set_ylabel(y_label)
            ax.legend()

    return axes


def plot_histograms_of_interval_distributions(
    model,
    sample,
    semantics=None,
    plot_pdf=True,
    max_cols=4,
):
    """
    Plot histograms of all model dimensions.

    If the model is conditional in a dimension all histograms per interval are plotted for that dimension.
    In such a case the fitted interval distributions are generally different from the final joint distribution.

    Parameters
    ----------
    model : MultivariateModel
        The model that was fitted to the dataset.
    sample : ndarray
        The data that was used to fit the model.
    semantics: dict, optional
        The description of the model. If None (the default) generic semantics will be used.
    plot_pdf: boolean, optional
        Whether the fitted probability density should be plotted. Defaults to True.
    max_cols : int, default 4
        Maximum number of columns (subfigures horizontally).

    Returns
    -------
    The used matplotlib axes objects.

    """
    sample = np.asarray(sample)
    n_dim = model.n_dim

    if semantics is None:
        semantics = get_default_semantics(n_dim)

    figures = []
    axes_list = []

    for dim in range(n_dim):
        x_name = semantics["names"][dim]
        x_symbol = semantics["symbols"][dim]
        x_unit = semantics["units"][dim]
        x_label = f"{x_name}," + r" $\it{" + f"{x_symbol}" + r"}$" + f" ({x_unit})"

        if model.conditional_on[dim] is None:
            # unconditional
            data = sample[:, dim]
            dist = model.distributions[dim]
            x = np.linspace(np.min(data), np.max(data))
            fig, ax = plt.subplots()
            ax.hist(
                data,
                bins="doane",
                density=True,
                color="#000000",
                histtype="stepfilled",
                alpha=0.2,
            )
            if plot_pdf:
                density = dist.pdf(x)
                ax.plot(
                    x,
                    density,
                    color="#004488",
                )
            ax.set_xlabel(x_label)
            ax.set_ylabel("Probability density")
            ax.set_title(f"n={len(data)}")

            figures.append(fig)
            axes_list.append(ax)

        else:
            # conditional
            cond_dist = model.distributions[dim]
            dist_per_interval = cond_dist.distributions_per_interval
            n_intervals = len(dist_per_interval)

            conditioning_idx = model.conditional_on[dim]
            conditioning_symbol = semantics["symbols"][conditioning_idx]
            conditioning_unit = semantics["units"][conditioning_idx]

            slicer = model.interval_slicers[conditioning_idx]
            interval_slices, conditioning_values, _ = slicer.slice_(
                sample[:, conditioning_idx]
            )
            data_intervals = [sample[int_slice, dim] for int_slice in interval_slices]

            # if the following fails, the sample was probably not used for fitting
            # TODO raise proper exception then
            np.testing.assert_allclose(
                conditioning_values, cond_dist.conditioning_values
            )
            assert len(data_intervals) == len(cond_dist.data_intervals)
            for i in range(len(data_intervals)):
                np.testing.assert_allclose(
                    np.sort(data_intervals[i]), np.sort(cond_dist.data_intervals[i])
                )

            fig, axes = _get_n_axes(n_intervals, max_cols=max_cols)
            for interval_idx in range(n_intervals):
                cond_val = conditioning_values[interval_idx]
                data = data_intervals[interval_idx]
                x = np.linspace(np.min(data), np.max(data))
                dist = dist_per_interval[interval_idx]
                ax = axes[interval_idx]
                ax.hist(
                    data,
                    bins="doane",
                    density=True,
                    color="#000000",
                    histtype="stepfilled",
                    alpha=0.2,
                )
                if plot_pdf:
                    density = dist.pdf(x)
                    ax.plot(
                        x,
                        density,
                        color="#004488",
                    )
                ax.set_xlabel(x_label)
                ax.set_ylabel("Probability density")
                title = f"{conditioning_symbol} = {cond_val:.3f} {conditioning_unit}, n={len(data)}"
                ax.set_title(title)

            figures.append(fig)
            axes_list.append(axes)

    return figures, axes_list


def plot_2D_isodensity(
    model,
    sample,
    semantics=None,
    swap_axis=False,
    limits=None,
    levels=None,
    ax=None,
    n_grid_steps=250,
    cmap=None,
    contour_kwargs=None,
):
    """
    Plot isodensity contours and a data sample for a 2D model.

    Parameters
    ----------
    model :  MultivariateModel
        The 2D model to use.
    sample : ndarray of floats
        The 2D data sample that should be plotted.
    semantics: dict, optional
        The description of the model. If None (the default) generic semantics
        will be used.
    swap_axis : boolean, optional
        If True the second dimension of the model is plotted on the x-axis and
        the first on the y-axis. Otherwise, vice-versa. Defaults to False.
    limits : list of tuples, optional
        Specifies in which rectangular region the density is calculated and
        plotted. If None (default) limits will be set automatically.
        Example: [(0, 20), (0, 12)]
    levels : list of floats, optional
        The probability density levels that are plotted. If None (default)
        levels are set automatically.
        Example: [0.001, 0.01, 0.1]
    n_grid_steps : int, optional
        The number of steps along each axis of the grid used to plot the contours.
        Defaults to 250.
    ax : matplotlib Axes, optional
        Matplotlib axes object to use for plotting. If None (default) a new
        figure will be created.
    cmap : matplotlib colormap, optional
        The colormap to use for the isodensity contours.
    contour_kwargs : dict
        Any other keyword arguments for matplotlib.pyplot.contour().
        Example: {"linewidths": [1,2,3], "linestyles": ["solid","dotted","dashed"]}

    Returns
    -------
    The used matplotlib axes object.

    """

    n_dim = model.n_dim
    assert n_dim == 2

    if swap_axis:
        x_idx = 1
        y_idx = 0
    else:
        x_idx = 0
        y_idx = 1

    if semantics is None:
        semantics = get_default_semantics(n_dim)

    if ax is None:
        _, ax = plt.subplots()

    sample = np.asarray(sample)
    ax.scatter(
        sample[:, x_idx],
        sample[:, y_idx],
        c="k",
        marker=".",
        alpha=0.3,
        rasterized=True,
    )

    if limits is not None:
        x_lower = limits[0][0]
        x_upper = limits[0][1]
        y_lower = limits[1][0]
        y_upper = limits[1][1]
    else:
        x_range = max(sample[:, 0]) - min((sample[:, 0]))
        expand_factor = 0.05
        x_lower = min(sample[:, 0]) - expand_factor * x_range
        x_upper = max(sample[:, 0]) + expand_factor * x_range
        y_range = max(sample[:, 1]) - min((sample[:, 1]))
        y_lower = min(sample[:, 1]) - expand_factor * y_range
        y_upper = max(sample[:, 1]) + expand_factor * y_range

    x, y = np.linspace((x_lower, y_lower), (x_upper, y_upper), num=n_grid_steps).T
    X, Y = np.meshgrid(x, y)
    grid_flat = np.c_[X.ravel(), Y.ravel()]
    f = model.pdf(grid_flat)
    Z = f.reshape(X.shape)

    if swap_axis:
        tmp = X
        X = Y
        Y = tmp

    if levels is not None:
        lvl_labels = ["{:.1E}".format(x) for x in levels]
        n_levels = len(levels)
    else:
        # Define the lowest isodensity level based on the density values on the evaluated grid.
        q = np.quantile(f, q=0.5)
        if q > 0:
            min_lvl = int(f"{q:.0e}".split("e")[1])
        else:
            min_lvl = -5
        n_levels = np.abs(min_lvl)
        levels = np.logspace(-1, min_lvl, num=n_levels)[::-1]
        lvl_labels = [f"1E{int(i)}" for i in np.linspace(-1, min_lvl, num=n_levels)][
            ::-1
        ]

    if cmap is None:
        cmap = _rainbow_PuRd()
    elif isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    colors = cmap(np.linspace(0, 1, num=n_levels))
    contour_kwargs = {} if contour_kwargs is None else contour_kwargs
    CS = ax.contour(X, Y, Z, levels=levels, colors=colors, **contour_kwargs)
    handles, _ = CS.legend_elements()
    ax.legend(
        handles=handles,
        labels=lvl_labels,
        loc="upper left",
        ncol=1,
        frameon=False,
        title="Probability density",
    )
    x_name = semantics["names"][x_idx]
    x_symbol = semantics["symbols"][x_idx]
    x_unit = semantics["units"][x_idx]
    x_label = f"{x_name}," + r" $\it{" + f"{x_symbol}" + r"}$" + f" ({x_unit})"
    y_name = semantics["names"][y_idx]
    y_symbol = semantics["symbols"][y_idx]
    y_unit = semantics["units"][y_idx]
    y_label = f"{y_name}," + r" $\it{" + f"{y_symbol}" + r"}$" + f" ({y_unit})"
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    return ax


def plot_2D_contour(
    contour,
    sample=None,
    design_conditions=None,
    semantics=None,
    swap_axis=False,
    ax=None,
    plot_kwargs=None,
):
    """
    Plot a 2D contour.

    Parameters
    ----------
    contour: Contour
        The 2D contour that should be plotted.
    sample : ndarray of floats, optional
        A 2D data sample that should be plotted along the contour.
        Shape: (number of realizations, 2)
    design_conditions : array-like or boolean, optional
       Specified environmental conditions under which the system must operate.
       If an array it is assumed to be a 2D array of shape
       (number of points, 2) and should contain the precalculated design
       conditions. If it is True design_conditions are computed with default
       values and plotted. Otherwise, no design conditions will be plotted
       (the default).
    semantics: dict, optional
        Generated model description. Defaults to None.
    swap_axis : boolean, optional
        f True the second dimension of the model is plotted on the x-axis and
        the first on the y-axis. Otherwise, vice-versa. Defaults to False.
    ax : matplotlib Axes, optional
        Matplotlib axes object to use for plotting. If None (default) a new
        figure will be created.
    plot_kwargs : dict
        Any other keyword arguments to pass to the matplotlib plotting function,
        Example: {"color": "blue", "linewidth": 2, "label": "100-year contour"

    Returns
    -------
    The matplotlib axes objects plotted into.
    (optional: the design_conditions if not None, yet to implement)

    """

    # design conditions can be True or array
    n_dim = 2
    if swap_axis:
        x_idx = 1
        y_idx = 0
    else:
        x_idx = 0
        y_idx = 1

    if semantics is None:
        semantics = get_default_semantics(n_dim)

    if ax is None:
        _, ax = plt.subplots()

    if design_conditions:
        try:  # if iterable assume it's already the design conditions
            iter(design_conditions)
        except (
            TypeError
        ):  # if it is not an array we compute the default design_conditions
            design_conditions = calculate_design_conditions(
                contour, swap_axis=swap_axis
            )

        ax.scatter(
            design_conditions[:, 0],
            design_conditions[:, 1],
            c="#DDAA33",
            marker="x",
            zorder=2.5,
        )

    if sample is not None:
        sample = np.asarray(sample)
        ax.scatter(
            sample[:, x_idx],
            sample[:, y_idx],
            c="k",
            marker=".",
            alpha=0.3,
            rasterized=True,
        )

    coords = contour.coordinates
    x = coords[:, x_idx].tolist()
    x.append(x[0])
    y = coords[:, y_idx].tolist()
    y.append(y[0])

    # It was thought that this line caused a DepreciationWarning, but the change
    # was reverted as we were not sure about the reason.
    # https://github.com/virocon-organization/virocon/commit/45482e0b5ff2d21c594f0e292b3db9c971881b5c
    # https://github.com/virocon-organization/virocon/pull/124#discussion_r684193507
    plot_kwargs = {} if plot_kwargs is None else plot_kwargs
    if "color" not in plot_kwargs and "c" not in plot_kwargs:
        plot_kwargs["c"] = "#BB5566"
    ax.plot(x, y, **plot_kwargs)
    # ax.plot(np.asarray(x, dtype=object), np.asarray(y, dtype=object), c="#BB5566")

    x_name = semantics["names"][x_idx]
    x_symbol = semantics["symbols"][x_idx]
    x_unit = semantics["units"][x_idx]
    x_label = f"{x_name}," + r" $\it{" + f"{x_symbol}" + r"}$" + f" ({x_unit})"
    y_name = semantics["names"][y_idx]
    y_symbol = semantics["symbols"][y_idx]
    y_unit = semantics["units"][y_idx]
    y_label = f"{y_name}," + r" $\it{" + f"{y_symbol}" + r"}$" + f" ({y_unit})"
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if design_conditions is None:
        return ax
    else:
        return ax, design_conditions



import numpy as np

from virocon._intersection import intersection


def test_intersection():
    x = np.linspace(1, 5, 11)
    y1 = 0 * x + 1
    y2 = 0.5 * x

    # The curves intersect at point (2, 1).
    ix, iy = intersection(x, y1, x, y2)

    np.testing.assert_allclose(ix, 2, atol=0.001)
    np.testing.assert_allclose(iy, 1, atol=0.001)



"""
Classes for environmental contours.
"""

import os
import warnings

import numpy as np
import scipy.stats as sts
import scipy.ndimage as ndi

from abc import ABC, abstractmethod

from virocon._nsphere import NSphere
from virocon.plotting import get_default_semantics
from virocon.utils import sort_points_to_form_continuous_line
from virocon.jointmodels import GlobalHierarchicalModel, TransformedModel

__all__ = [
    "calculate_alpha",
    "save_contour_coordinates",
    "IFORMContour",
    "ISORMContour",
    "HighestDensityContour",
    "DirectSamplingContour",
    "AndContour",
    "OrContour",
]


def calculate_alpha(state_duration, return_period):
    """
    Calculates the probability that an environmental contour is exceeded
    (exceedance probability).

    The exceedance probability, α, corresponds to a certain recurrence or
    return period, T, which describes the average time period between two
    consecutive environmental states that exceed the contour . Note that
    exceedance can be defined in various ways for environmental contours
    (Mackay and Haselsteiner, 2021) [1]_

    Parameters
    ----------
    state_duration : float
        Time period for which an environmental state is measured,
        expressed in hours :math:`(T_s)`.
    return_period : float
        Describes the average time period between two consecutive
        environmental states that exceed a contour.

    Returns
    -------
    alpha : float
        The probability that an environmental contour is exceeded.

    References
    ----------
    .. [1] Mackay, E., & Haselsteiner, A. F. (2021).
       Marginal and total exceedance probabilities of environmental contours.
       Marine Structures, 75. https://doi.org/10.1016/j.marstruc.2020.102863

    """

    alpha = state_duration / (return_period * 365.25 * 24)
    return alpha


def save_contour_coordinates(contour, file_path, semantics=None):
    """
    Saves the coordinates of the calculated contour.
    Saves a .txt file to the given path.

    Parameters
    ----------
    contour : Contour
     The contour with the coordinates to save.
    file_path : string
     Indicates the path, where the contour coordinates are saved.
    semantics : dictionary
     The description of the model. semantics has the keys 'names', 'symbols'
     and 'units'. Each value is a list of strings. For each dimension of the
     model the strings describe the name, symbol or unit of that dimension,
     respectively. This information is used as the header of the created file.
     Defaults to a dict with general descriptions.

    """

    root, ext = os.path.splitext(file_path)
    if not ext:
        file_path += ".txt"

    n_dim = contour.coordinates.shape[1]
    if semantics is None:
        semantics = get_default_semantics(n_dim)

    header = ";".join(
        (f"{semantics['names'][d]} ({semantics['units'][d]})" for d in range(n_dim))
    )

    np.savetxt(
        file_path,
        contour.coordinates,
        fmt="%1.6f",
        delimiter=";",
        header=header,
        comments="",
    )


class Contour(ABC):
    """
      Abstract base class for contours.

      A contour implements a method to define multivariate extremes based on a
      joint probabilistic model of variables like significant wave height,
      wind speed or spectral peak period.

      Contour curves or surfaces for more than two environmental parameters
      give combination of environmental parameters which approximately
      describe the various actions corresponding to the given exceedance
      probability [2]_.


    References
    ----------
    .. [2] NORSOK standard N-003, Edition 2, September 2007. Actions and
        action effects.


    """

    def __init__(self):
        try:
            _ = self.model
        except AttributeError:
            raise NotImplementedError(
                f"Can't instantiate abstract class {type(self).__name__} "
                "with abstract attribute model."
            )
        try:
            _ = self.alpha
        except AttributeError:
            raise NotImplementedError(
                f"Can't instantiate abstract class {type(self).__name__} "
                "with abstract attribute model."
            )

        self._compute()
        try:
            _ = self.coordinates
        except AttributeError:
            raise NotImplementedError(
                f"Can't instantiate abstract class {type(self).__name__} "
                "with abstract attribute coordinates."
            )

    @abstractmethod
    def _compute(self):
        """
        Compute the contours coordinates.

        Is automatically called in the __init__.
        """
        pass


class IFORMContour(Contour):
    """
    Contour based on the inverse first-order reliability method.

    This method was proposed by Winterstein et. al (1993) [3]_

    Parameters
    ----------
    model :  MultivariateModel | TransformedModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation falls
        outside the environmental contour.
    n_points : int, optional
        Number of points on the contour. Defaults to 180.

    Attributes
    ----------
    coordinates :
        Coordinates of the calculated contour.
    beta :
        Reliability index.
    sphere_points :
          Points of the sphere in U space [3]_ .

    References
    ----------
    .. [3] Winterstein, S.R.; Ude, T.C.; Cornell, C.A.; Bjerager, P.; Haver, S. (1993)
        Environmental parameters  for extreme response: Inverse FORM with omission
        factors. ICOSSAR 93, Innsbruck, Austria.


    """

    def __init__(self, model, alpha, n_points=180):
        allowed_model_types = (GlobalHierarchicalModel, TransformedModel)
        if isinstance(model, allowed_model_types):
            self.model = model
        else:
            raise TypeError(
                f"Type of model was {type(model).__name__} but expected one of {allowed_model_types}"
            )
        self.alpha = alpha
        self.n_points = n_points
        super().__init__()

    def _compute(
        self,
    ):
        """
        Calculates coordinates using IFORM.

        """
        n_dim = self.model.n_dim
        n_points = self.n_points

        # A GlobalHierarchicalModel has the attributes distributions and conditional_on
        # but a TransformedModel not. Consequently, contour calculation requries different
        # algorithms for these two cases. TransformedModel is used the EW models defined in
        # predefined.py .
        if isinstance(self.model, GlobalHierarchicalModel):
            distributions = self.model.distributions
            conditional_on = self.model.conditional_on
        elif isinstance(self.model, TransformedModel):
            distributions = None
            conditional_on = None
        else:
            raise TypeError()

        beta = sts.norm.ppf(1 - self.alpha)
        self.beta = beta

        # TODO Update NSphere to handle n_dim case with order
        # Create sphere
        if n_dim == 2:
            _phi = np.linspace(0, 2 * np.pi, num=n_points, endpoint=False)
            _x = np.cos(_phi)
            _y = np.sin(_phi)
            _circle = np.stack((_x, _y), axis=1)
            sphere_points = beta * _circle

        else:
            sphere = NSphere(dim=n_dim, n_samples=n_points)
            sphere_points = beta * sphere.unit_sphere_points

        # Get probabilities for coordinates
        norm_cdf = sts.norm.cdf(sphere_points)

        # Inverse procedure. Get coordinates from probabilities.
        p = norm_cdf
        coordinates = np.empty_like(p)

        if distributions:
            coordinates[:, 0] = distributions[0].icdf(p[:, 0])
        else:
            coordinates[:, 0] = self.model.marginal_icdf(
                p[:, 0], 0, precision_factor=self.model.precision_factor
            )

        for i in range(1, n_dim):
            if distributions:
                if conditional_on[i] is None:
                    coordinates[:, i] = distributions[i].icdf(p[:, i])
                else:
                    cond_idx = conditional_on[i]
                    coordinates[:, i] = distributions[i].icdf(
                        p[:, i], given=coordinates[:, cond_idx]
                    )
            else:
                given = coordinates[:, np.arange(n_dim) != i]
                coordinates[:, i] = self.model.conditional_icdf(
                    p[:, i], i, given, random_state=self.model.random_state
                )

        self.sphere_points = sphere_points
        self.coordinates = coordinates


class ISORMContour(Contour):
    """
    Contour based on the inverse second-order reliability method.

    This method was proposed by Chai and Leira (2018) [4]_

    Parameters
    ----------
    model : MultivariateModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation falls
        outside the environmental contour.
    n_points : int, optional
        Number of points on the contour. Defaults to 180.

    Attributes
    ----------
    coordinates :
        Coordinates of the calculated contour.
    beta :
        Reliability index.
    sphere_points :
          Points of the sphere in U space [4]_ .

    References
    ----------
    .. [4] Chai, W.; Leira, B.J. (2018)
        Environmental contours based on inverse SORM. Marine Structures Volume 60,
        pp. 34-51. DOI: 10.1016/j.marstruc.2018.03.007 .

    """

    def __init__(self, model, alpha, n_points=180):
        self.model = model
        self.alpha = alpha
        self.n_points = n_points
        super().__init__()

    def _compute(
        self,
    ):
        """
        Calculates coordinates using ISORM.

        """

        n_dim = self.model.n_dim
        n_points = self.n_points

        distributions = self.model.distributions
        conditional_on = self.model.conditional_on

        # Use the ICDF of a chi-squared distribution with n dimensions. For
        # reference see equation 20 in Chai and Leira (2018).
        beta = np.sqrt(sts.chi2.ppf(1 - self.alpha, n_dim))

        # Create sphere.
        if n_dim == 2:
            _phi = np.linspace(0, 2 * np.pi, num=n_points, endpoint=False)
            _x = np.cos(_phi)
            _y = np.sin(_phi)
            _circle = np.stack((_x, _y)).T
            sphere_points = beta * _circle

        else:
            sphere = NSphere(dim=n_dim, n_samples=n_points)
            sphere_points = beta * sphere.unit_sphere_points

        # Get probabilities for coordinates of shape.
        norm_cdf_per_dimension = [
            sts.norm.cdf(sphere_points[:, dim]) for dim in range(n_dim)
        ]

        # Inverse procedure. Get coordinates from probabilities.
        data = np.zeros((n_points, n_dim))

        for i in range(n_dim):
            dist = distributions[i]
            cond_idx = conditional_on[i]
            if cond_idx is None:
                data[:, i] = dist.icdf(norm_cdf_per_dimension[i])
            else:
                conditioning_values = data[:, cond_idx]
                for j in range(n_points):
                    data[j, i] = dist.icdf(
                        norm_cdf_per_dimension[i][j], given=conditioning_values[j]
                    )

        self.beta = beta
        self.sphere_points = sphere_points
        self.coordinates = data


class HighestDensityContour(Contour):
    """
    Contour based on the highest density method.

    This method was proposed by Haselsteiner et. al (2017) [5]_

    Parameters
    ----------
    model : MultivariateModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation
        falls outside the environmental contour.
    limits : list of tuples, optional
       The limits of the grid to use for calculation. One 2-element tuple
       for each dimension of the model, containing the minimum and maximum
       for that dimension. (min, max). If not given, reasonable values are
       choosen using the models marginal_icdf as upper limit and 0 as lower
       limit.
    deltas : float or list of float, optional
       The step size of the grid to use for calculation. If a single float
       is supplied it is used for all dimensions. If a list is supplied
       there has to be one entry for each dimension of the model. Defaults
       to 0.25% of the range defined by limits.

    Attributes
    ----------
    coordinates : ndarray
        Coordinates of the calculated contour.
        Shape: (number of points, number of dimensions).
    cell_center_coordinates : list of array
        Points at which the grid is evaluated.
        A list with one entry for each dimension, each entry is an array with
        the cell centers for that dimension.
    fm : float
        Minimum probability density of the enclosed region / constant
        probability density along the contour.


    References
    ----------
    .. [5] Haselsteiner, A.F.; Ohlendorf, J.H.; Wosniok, W.; Thoben, K.D. (2017)
        Deriving environmental contours from highest density regions,
        Coastal Engineering, Volume 123. DOI: 10.1016/j.coastaleng.2017.03.002.

    """

    def __init__(self, model, alpha, limits=None, deltas=None):
        self.model = model
        self.alpha = alpha
        self.limits = limits
        self.deltas = deltas
        self._check_grid()
        super().__init__()

    def _check_grid(self):
        n_dim = self.model.n_dim
        limits = self.limits
        deltas = self.deltas

        if limits is None:
            alpha = self.alpha
            marginal_icdf = self.model.marginal_icdf
            non_exceedance_p = 1 - 0.2**n_dim * alpha
            limits = [
                (0, marginal_icdf(non_exceedance_p, dim, precision_factor=0.05))
                for dim in range(n_dim)
            ]
            # TODO use distributions lower bound instead of zero
        else:
            # Check limits length.
            if len(limits) != n_dim:
                raise ValueError(
                    "limits has to be of length equal to number of dimensions, "
                    f"but len(limits)={len(limits)}, n_dim={n_dim}."
                )
        self.limits = limits

        if deltas is None:
            deltas = np.empty(shape=n_dim)
            # Set default cell size to 0.25 percent of the variable space.
            # This is losely based on the results from Fig. 7 in 10.1016/j.coastaleng.2017.03.002
            # In the considered variable space length of 20 a cell length of
            # 0.05 was sufficient --> 20 / 0.05 = 400. 1/400 = 0.0025
            relative_cell_size = 0.0025

            for i in range(n_dim):
                deltas[i] = (limits[i][1] - limits[i][0]) * relative_cell_size
        else:
            try:  # Check if deltas is an iterable
                iter(deltas)
                if len(deltas) != n_dim:
                    raise ValueError(
                        "deltas has do be either scalar, "
                        "or list of length equal to number of dimensions, "
                        f"but was list of length {len(deltas)}"
                    )
                deltas = list(deltas)
            except TypeError:  # asserts that deltas is scalar
                deltas = [deltas] * n_dim

        self.deltas = deltas

    def _compute(self):
        """
        Calculates coordinates using HDC.

        """

        limits = self.limits
        deltas = self.deltas
        n_dim = self.model.n_dim
        alpha = self.alpha

        # Create sampling coordinate arrays.
        cell_center_coordinates = []
        for i, lim_tuple in enumerate(limits):
            try:
                iter(lim_tuple)
                if len(lim_tuple) != 2:
                    raise ValueError(
                        "tuples in limits have to be of length 2 ( = (min, max)), "
                        f"but tuple with index = {i}, has length = {len(lim_tuple)}."
                    )

            except TypeError:
                raise ValueError(
                    "tuples in limits have to be of length 2 ( = (min, max)), "
                    f"but tuple with index = {i}, has length = 1."
                )

            min_ = min(lim_tuple)
            max_ = max(lim_tuple)
            delta = deltas[i]
            samples = np.arange(min_, max_ + delta, delta)
            cell_center_coordinates.append(samples)

        f = self.cell_averaged_joint_pdf(cell_center_coordinates)  # TODO

        if np.isnan(f).any():
            raise ValueError(
                "Encountered nan in cell averaged probabilty joint pdf. "
                "Possibly invalid distribution parameters?"
            )

        # Calculate probability per cell.
        cell_prob = f
        for delta in deltas:
            cell_prob *= delta

        # Calculate highest density region.
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                HDR, prob_m = self.cumsum_biggest_until(cell_prob, 1 - alpha)
        except RuntimeWarning:
            HDR = np.ones_like(cell_prob)
            prob_m = 0
            warnings.warn(
                "A probability of 1-alpha could not be reached. "
                "Consider enlarging the area defined by limits or "
                "setting n_years to a smaller value.",
                RuntimeWarning,
                stacklevel=4,
            )

        # Calculate fm from probability per cell.
        fm = prob_m
        for delta in deltas:
            fm /= delta

        structure = np.ones(tuple([3] * n_dim), dtype=bool)
        HDC = HDR - ndi.binary_erosion(HDR, structure=structure)

        labeled_array, n_modes = ndi.label(HDC, structure=structure)

        coordinates = []
        # Iterate over all partial contours and start at 1.
        for i in range(1, n_modes + 1):
            # Array of arrays with same length, one per dimension
            # containing the indice of the contour.
            partial_contour_indice = np.nonzero(labeled_array == i)

            # Calculate the values corresponding to the indice
            partial_coordinates = []
            for dimension, indice in enumerate(partial_contour_indice):
                partial_coordinates.append(cell_center_coordinates[dimension][indice])

            coordinates.append(partial_coordinates)

        is_single_contour = False
        if len(coordinates) == 1:
            is_single_contour = True
            coordinates = coordinates[0]

        self.cell_center_coordinates = cell_center_coordinates
        self.fm = fm

        if is_single_contour:
            if n_dim == 2:
                self.coordinates = np.array(
                    sort_points_to_form_continuous_line(
                        *coordinates, search_for_optimal_start=True
                    )
                ).T
            else:
                self.coordinates = np.array(coordinates).T
        else:
            self.coordinates = coordinates
            # TODO raise warning

    @staticmethod
    def cumsum_biggest_until(array, limit):
        """
        Find biggest elements to sum to reach limit.

        Sorts array, and calculates the cumulative sum.
        Returns a boolean array with the same shape as array indicating the
        fields summed to reach limit, as well as the last value added.

        Parameters
        ----------
        array : ndarray
            Array of arbitrary shape with all values >= 0.
        limit : float
            Value to sum up to.

        Returns
        -------
        summed_fields : ndarray, dtype=Bool
            Boolean array of shape like array with True if element was used in summation.
        last_summed : float
            Element that was added last to the sum.

        Raises
        ------
        ValueError
            If `array` contains nan.
        Notes
        ------
        A ``RuntimeWarning`` is raised if the limit cannot be reached by summing all values.

        """

        flat_array = np.ravel(array)
        if np.isnan(flat_array).any():
            raise ValueError("array contains nan.")

        sort_inds = np.argsort(flat_array, kind="mergesort")[::-1]
        sort_vals = flat_array[sort_inds]

        cum_sum = np.cumsum(sort_vals)

        if cum_sum[-1] < limit:
            warnings.warn(
                "The limit could not be reached.", RuntimeWarning, stacklevel=1
            )

        summed_flat_inds = sort_inds[cum_sum <= limit]

        summed_fields = np.zeros(array.shape)

        summed_fields[np.unravel_index(summed_flat_inds, shape=array.shape)] = 1

        last_summed = array[np.unravel_index(summed_flat_inds[-1], shape=array.shape)]

        return summed_fields, last_summed

    def cell_averaged_joint_pdf(self, coords):
        """
        Calculates the cell averaged joint probabilty density function.

        Multiplies the cell averaged probability densities of all distributions.

        Parameters
        ----------
        coords : list of array
            List with one coordinate array for each dimension.

        Returns
        -------
        fbar : array
            Joint cell averaged probability density function evaluated at coords.
            Cell averaged probability density function evaluated at coords.
            n dimensional array, where n is the number of dimensions of the
            model used for calculation.


        """

        n_dim = len(coords)
        fbar = np.ones(((1,) * n_dim), dtype=np.float64)
        for dist_idx in range(n_dim):
            fbar = np.multiply(fbar, self.cell_averaged_pdf(dist_idx, coords))

        return fbar

    def cell_averaged_pdf(self, dist_idx, coords):
        """
        Calculates the cell averaged probabilty density function of a single
        distribution.

        Calculates the pdf by approximating it with the finite differential
        quotient of the cumulative distributions function, evaluated at the
        grid cells borders.
        i.e. :math:`f(x) \\approx \\frac{F(x+ 0.5\\Delta x) - F(x- 0.5\\Delta x) }{\\Delta x}`

        Parameters
        ----------
        dist_idx : int
            The index of the distribution to calcululate the pdf for.

        coords : list of array
            List with one coordinate array for each dimension.

        Returns
        -------
        fbar : array
            Cell averaged probability density function evaluated at coords.
            n dimensional array, where n is the number of dimensions of the
            model used for calculation. All dimensions but, the dist_idx and
            the cond_idx dimensions are of length 1. The dist_idx and cond_idx
            dimensions are of length equal to the length of coords.


        """

        n_dim = len(coords)
        dist = self.model.distributions[dist_idx]
        cond_idx = self.model.conditional_on[dist_idx]

        dx = coords[dist_idx][1] - coords[dist_idx][0]

        cdf = dist.cdf
        fbar_out_shape = np.ones(n_dim, dtype=int)

        if cond_idx is None:  # independent variable
            # Calculate averaged pdf.
            lower = cdf(coords[dist_idx] - 0.5 * dx)
            upper = cdf(coords[dist_idx] + 0.5 * dx)
            fbar = upper - lower

            fbar_out_shape[dist_idx] = len(coords[dist_idx])

        else:
            dist_values = coords[dist_idx]
            cond_values = coords[cond_idx]
            fbar = np.empty((len(cond_values), len(dist_values)))
            for i, cond_value in enumerate(cond_values):
                lower = cdf(coords[dist_idx] - 0.5 * dx, given=cond_value)
                upper = cdf(coords[dist_idx] + 0.5 * dx, given=cond_value)
                fbar[i, :] = upper - lower

            fbar_out_shape[dist_idx] = len(coords[dist_idx])
            fbar_out_shape[cond_idx] = len(coords[cond_idx])

        fbar_out = fbar.reshape(fbar_out_shape)
        return fbar_out / dx


class DirectSamplingContour(Contour):
    """
    Direct sampling contour as introduced by Huseby et al. (2013) [6]_
    The provided direct sampling contour method only works for 2D models.

    Parameters
    ----------
    model : MultivariateModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation
        falls outside the environmental contour.
    n : int, optional
        Number of data points that shall be Monte Carlo simulated. Defaults
        to None, which calculates n based on alpha: n = int(100 / alpha).
    deg_step : float, optional
        Directional step in degrees. Defaults to 5.
    sample : 2-dimensional ndarray, optional
        Monte Carlo simulated environmental states. Array is of shape (n, d)
        with d being the number of variables and n being the number of
        observations.

    Attributes
    ----------
    coordinates : ndarray
        Coordinates of the calculated contour.
        Shape: (number of points, number of dimensions).

    References
    ----------
    .. [6] Huseby, A.B.; Vanem, E.; Natvig, B. (2013)
        A new approach to environmental contours for ocean engineering applications
        based on direct Monte Carlo simulations,
        Ocean Engineering, Volume 60. DOI: doi.org/10.1016/j.oceaneng.2012.12.034

    """

    def __init__(self, model, alpha, n=None, deg_step=5, sample=None):
        self.model = model
        self.alpha = alpha
        if n is None:
            n = int(100 / alpha)
        self.n = n
        self.deg_step = deg_step
        self.sample = sample
        super().__init__()

    def _compute(self):
        sample = self.sample
        n = self.n
        deg_step = self.deg_step
        alpha = self.alpha

        if self.model.n_dim != 2:
            raise NotImplementedError(
                "DirectSamplingContour is currently only "
                "implemented for two dimensions."
            )

        if sample is None:
            sample = self.model.draw_sample(n)
            self.sample = sample
        x, y = sample.T

        # Calculate non-exceedance probability.
        # alpha = 1 - (1 / (self.return_period * 365.25 * 24 / self.state_duration))
        non_exceedance_p = 1 - alpha

        # Define the angles such the coordinates[0] and coordinates[1] will
        # be based on the exceedance plane with angle 0 deg, with 0 deg being
        # along the x-axis. Angles will increase counterclockwise in a xy-plot.
        # Not enirely sure why the + 2*rad_step is required, but tests show it.
        rad_step = deg_step * np.pi / 180
        angles = np.arange(
            0.5 * np.pi + 2 * rad_step, -1.5 * np.pi + rad_step, -1 * rad_step
        )

        length_t = len(angles)
        r = np.zeros(length_t)

        # Find radius for each angle.
        i = 0
        while i < length_t:
            z = x * np.cos(angles[i]) + y * np.sin(angles[i])
            r[i] = np.quantile(z, non_exceedance_p)
            i = i + 1

        # Find intersection of lines.
        a = np.array(np.concatenate((angles, [angles[0]]), axis=0))
        r = np.array(np.concatenate((r, [r[0]]), axis=0))

        denominator = np.sin(a[2:]) * np.cos(a[1 : len(a) - 1]) - np.sin(
            a[1 : len(a) - 1]
        ) * np.cos(a[2:])

        x_cont = (
            np.sin(a[2:]) * r[1 : len(r) - 1] - np.sin(a[1 : len(a) - 1]) * r[2:]
        ) / denominator
        y_cont = (
            -np.cos(a[2:]) * r[1 : len(r) - 1] + np.cos(a[1 : len(a) - 1]) * r[2:]
        ) / denominator

        self.coordinates = np.array([x_cont, y_cont]).T


class AndContour(Contour):
    """
    A contour that connects points of constant AND exceedance. Such
    contours are described, for example, in Mazas (2019) [7]_.
    This implementation uses Monte Carlo simulation and only works for 2D models.

    Parameters
    ----------
    model : MultivariateModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation
        falls outside the environmental contour.
    n : int, optional
        Number of data points that shall be Monte Carlo simulated. Defaults
        to None, which calculates n based on alpha: n = int(100 / alpha).
    deg_step : float, optional
        Directional step in degrees. Defaults to 5.
    sample : 2-dimensional ndarray, optional
        Monte Carlo simulated environmental states. Array is of shape (n, d)
        with d being the number of variables and n being the number of
        observations.
    allowed_error : float, optional
        Required precision for the alpha value. For example 0.1 means that
        the algorithm searches along the path until the probability of exceedance
        at the current point p_e satisfies \| p_e - alpha \| / alpha < 0.1.
        Defaults to 0.01.

    Attributes
    ----------
    coordinates : ndarray
        Coordinates of the calculated contour.
        Shape: (number of points, number of dimensions).

    References
    ----------
    .. [7] Mazas, F. (2019). Extreme events: a framework for assessing natural
           hazards. Natural Hazards. https://doi.org/10.1007/s11069-019-03581-9

    """

    def __init__(
        self, model, alpha, n=None, deg_step=3, sample=None, allowed_error=0.01
    ):
        self.model = model
        self.alpha = alpha
        if n is None:
            n = int(100 / alpha)
        self.n = n
        self.deg_step = deg_step
        self.sample = sample
        self.allowed_error = allowed_error
        super().__init__()

    def _compute(self):
        model = self.model
        alpha = self.alpha
        n = self.n
        deg_step = self.deg_step
        sample = self.sample
        allowed_error = self.allowed_error

        if self.model.n_dim != 2:
            raise NotImplementedError(
                "AndContour is currently only implemented for two dimensions."
            )

        if sample is None:
            sample = self.model.draw_sample(n)
            self.sample = sample
        x, y = sample.T

        # This value was set empirically, no systematic testing has been performed.
        max_iterations = 100

        x_marginal = model.marginal_icdf(1 - alpha, 0)
        y_marginal = model.marginal_icdf(1 - alpha, 1)
        thetas = np.arange(0, 90, deg_step)
        coords_x = np.empty(thetas.size + 1)
        coords_y = np.empty(thetas.size + 1)

        # The algorithm works by moving along a line of angle theta relative
        # to the x-axis until the AND exceedance is alpha.
        for i, theta in enumerate(thetas):
            unity_vector = np.empty((2, 1))
            unity_vector[0] = np.cos(theta / 180 * np.pi)
            unity_vector[1] = np.sin(theta / 180 * np.pi)
            max_distance = np.sqrt(x_marginal**2 + y_marginal**2)
            rel_dist = 0.2
            rel_step_size = 0.1
            current_pe = 0  # pe = probability of exceedance.
            nr_iterations = 0
            while np.abs((current_pe - alpha)) / alpha > allowed_error:
                abs_dist = rel_dist * max_distance
                current_vector = unity_vector * abs_dist
                both_greater = np.logical_and(
                    x > current_vector[0], y > current_vector[1]
                )
                current_pe = both_greater.sum() / both_greater.size
                if current_pe > alpha:
                    rel_dist = rel_dist + rel_step_size
                else:
                    rel_step_size = 0.5 * rel_step_size
                    rel_dist = rel_dist - rel_step_size
                nr_iterations = nr_iterations + 1
                if nr_iterations == max_iterations:
                    warnings.warn(
                        "Could not achieve the required precision. Stopping "
                        "because the maximum number of iterations is reached.",
                        UserWarning,
                    )
                    break
            coords_x[i] = current_vector[0]
            coords_y[i] = current_vector[1]
        coords_x[-1] = 0
        coords_y[-1] = 0
        self.coordinates = np.array([coords_x, coords_y]).T


class OrContour(Contour):
    """
    A contour that connects points of constant OR exceedance. Such
    type of multivariate exceedance is described, for example, in Serinaldi (2015) [8]_.
    This implementation uses Monte Carlo simulation and only works for 2D models.

    Parameters
    ----------
    model : MultivariateModel
        The model to be used to calculate the contour.
    alpha : float
        The exceedance probability. The probability that an observation
        falls outside the environmental contour.
    n : int, optional
        Number of data points that shall be Monte Carlo simulated. Defaults
        to None, which calculates n based on alpha: n = int(100 / alpha).
    deg_step : float, optional
        Directional step in degrees. Defaults to 5.
    sample : 2-dimensional ndarray, optional
        Monte Carlo simulated environmental states. Array is of shape (n, d)
        with d being the number of variables and n being the number of
        observations.
    allowed_error : float, optional
        Required precision for the alpha value. For example 0.1 means that
        the algorithm searches along the path until the probability of exceedance
        at the current point p_e satisfies \| p_e - alpha \| / alpha < 0.1.
        Defaults to 0.01.
    lowest_theta : float, optional
        Lowest angle considered in the calculation of the contour. Given in deg.
        Defaults to 10.
    highest_theta : float, otptional
        Highest angle considered in the calculation of the contour. Given in deg.
        Defaults to 80.

    Attributes
    ----------
    coordinates : ndarray
        Coordinates of the calculated contour.
        Shape: (number of points, number of dimensions).

    References
    ----------
    .. [8] Serinaldi, F. (2015). Dismissing return periods! Stochastic Environmental
           Research and Risk Assessment, 29(4), 1179–1189. https://doi.org/10.1007/s00477-014-0916-1
    """

    def __init__(
        self,
        model,
        alpha,
        n=None,
        deg_step=3,
        sample=None,
        allowed_error=0.01,
        lowest_theta=10,
        highest_theta=80,
    ):
        self.model = model
        self.alpha = alpha
        if n is None:
            n = int(100 / alpha)
        self.n = n
        self.deg_step = deg_step
        self.sample = sample
        self.allowed_error = allowed_error
        self.lowest_theta = lowest_theta
        self.highest_theta = highest_theta
        super().__init__()

    def _compute(self):
        model = self.model
        alpha = self.alpha
        n = self.n
        deg_step = self.deg_step
        sample = self.sample
        allowed_error = self.allowed_error
        lowest_theta = self.lowest_theta
        highest_theta = self.highest_theta

        if self.model.n_dim != 2:
            raise NotImplementedError(
                "OrContour is currently only implemented for two dimensions."
            )

        if sample is None:
            sample = self.model.draw_sample(n)
            self.sample = sample
        x, y = sample.T

        # This value was set empirically, no systematic testing has been performed.
        max_iterations = 100

        x_marginal = model.marginal_icdf(1 - alpha, 0)
        y_marginal = model.marginal_icdf(1 - alpha, 1)
        max_factor = 1.1
        x_max_consider = max_factor * max(x)
        y_max_consider = max_factor * max(y)

        thetas = np.arange(lowest_theta, highest_theta, deg_step)

        coords_x = []
        coords_y = []

        # The algorithm works by moving along a line of angle theta relative
        # to the x-axis until the OR exceedance is alpha.
        for theta in thetas:
            unity_vector = np.empty((2, 1))
            unity_vector[0] = np.cos(theta / 180 * np.pi)
            unity_vector[1] = np.sin(theta / 180 * np.pi)
            max_distance = np.sqrt(x_marginal**2 + y_marginal**2)
            rel_dist = 0.2
            rel_step_size = 0.1
            current_pe = 0  # pe = probability of exceedance.
            nr_iterations = 0
            while np.abs((current_pe - alpha)) / alpha > allowed_error:
                abs_dist = rel_dist * max_distance
                current_vector = unity_vector * abs_dist
                or_exceeded = np.logical_or(
                    x > current_vector[0], y > current_vector[1]
                )
                current_pe = or_exceeded.sum() / or_exceeded.size
                if current_pe > alpha:
                    rel_dist = rel_dist + rel_step_size
                else:
                    rel_step_size = 0.5 * rel_step_size
                    rel_dist = rel_dist - rel_step_size
                nr_iterations = nr_iterations + 1
                if nr_iterations == max_iterations:
                    warnings.warn(
                        "Could not achieve the required precision. Stopping "
                        "because the maximum number of iterations is reached.",
                        UserWarning,
                    )
                    break
            if (current_vector[0] < x_max_consider) and (
                current_vector[1] < y_max_consider
            ):
                coords_x.append(current_vector[0])
                coords_y.append(current_vector[1])
        coords_x.append(0)
        coords_y.append(coords_y[-1])
        coords_x.append(0)
        coords_y.append(0)
        coords_x.append(coords_x[0])
        coords_y.append(0)

        coords_x = np.array(coords_x, dtype=object)
        coords_y = np.array(coords_y, dtype=object)

        self.coordinates = np.array([coords_x, coords_y]).T



"""
A comprehensive example that shows the whole workflow of
1) Loading data
2) Defining the model structure for a joint distribution
3) Estimating the parameter values of the model ("Fitting")
4) Computing an environmental contour

This example reproduces the results published in Haselsteiner
et al. (2019). The Hs-Tz joint distribution model recommended
in DNVGL-RP-C203 (2017) is used.

Haselsteiner, A. F., Coe, R. G., Manuel, L., Nguyen, P. T. T.,
Martin, N., & Eckert-Gallup, A. (2019). A benchmarking exercise
on estimating extreme environmental conditions: Methodology &
baseline results. Proc. 38th International Conference on Ocean,
Offshore and Arctic Engineering (OMAE 2019).
https://doi.org/10.1115/OMAE2019-96523

DNV GL. (2017). Recommended practice DNVGL-RP-C205:
Environmental conditions and environmental loads.
"""

import numpy as np
import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    WeibullDistribution,
    LogNormalDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    IFORMContour,
    plot_marginal_quantiles,
    plot_dependence_functions,
    plot_2D_contour,
)

# Load sea state measurements. This dataset has been used
# in a benchmarking exercise, see https://github.com/ec-benchmark-organizers/ec-benchmark
# The dataset was derived from NDBC buoy 44007, https://www.ndbc.noaa.gov/station_page.php?station=44007
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A.txt")

# Define the structure of the joint model that we will use to describe
# the the environmental data. To define a joint model, we define the
# univariate parametric distributions and the dependence structure.
# The dependence structure is defined using parametric functions.


# A 3-parameter power function, which will be used as a dependence function.
def _power3(x, a, b, c):
    return a + b * x**c


# A 3-parameter exponential function, which will be used a dependence function.
def _exp3(x, a, b, c):
    return a + b * np.exp(c * x)


# Lower and upper interval boundaries for the three parameter values.
bounds = [(0, None), (0, None), (None, None)]

power3 = DependenceFunction(_power3, bounds, latex="$a + b * x^c$")
exp3 = DependenceFunction(_exp3, bounds, latex="$a + b * \exp(c * x)$")

dist_description_0 = {
    "distribution": WeibullDistribution(),
    "intervals": WidthOfIntervalSlicer(width=0.5),
}
dist_description_1 = {
    "distribution": LogNormalDistribution(),
    "conditional_on": 0,
    "parameters": {"mu": power3, "sigma": exp3},
}
model = GlobalHierarchicalModel([dist_description_0, dist_description_1])

# Define a dictionary that describes the model.
semantics = {
    "names": ["Significant wave height", "Zero-up-crossing period"],
    "symbols": ["H_s", "T_z"],
    "units": ["m", "s"],
}

# Fit the model to the data (estimate the model's parameter values).
model.fit(data)

# Print the estimated parameter values
print(model)

# Create plots to inspect the model's goodness-of-fit.
fig1, axs = plt.subplots(1, 2, figsize=[10, 4.8], dpi=150)
plot_marginal_quantiles(model, data, semantics, axes=axs)
fig2, axs = plt.subplots(1, 2, figsize=[10, 4.8], dpi=150)
plot_dependence_functions(model, semantics, axes=axs)

# Compute an IFORM contour with a return period of 20 years.
state_duration = 1  # hours
return_period = 20  # years
alpha = state_duration / (return_period * 365.25 * 24)
contour = IFORMContour(model, alpha)

# Plot the contour on top of a scatter diagram of the metocean data.
ax = plot_2D_contour(contour, sample=data, semantics=semantics, swap_axis=True)
plt.show()



# ruff: noqa
__version__ = "2.4.0"
from virocon.contours import *
from virocon.dependencies import *
from virocon.distributions import *
from virocon.intervals import *
from virocon.jointmodels import *
from virocon.plotting import *
from virocon.predefined import *
from virocon.utils import *
from virocon.variable_transform import *



import pytest
import numpy as np

from virocon import (
    IFORMContour,
    ISORMContour,
    HighestDensityContour,
    DirectSamplingContour,
    AndContour,
    OrContour,
    calculate_alpha,
    DependenceFunction,
    LogNormalDistribution,
    WeibullDistribution,
    GlobalHierarchicalModel,
)


@pytest.fixture(scope="module")
def seastate_model():
    """
    This joint distribution model described by Vanem and Bitner-Gregersen (2012)
    is widely used in academia. Here, we use it for evaluation.
    DOI: 10.1016/j.apor.2012.05.006
    """

    def _power3(x, a=0.1000, b=1.489, c=0.1901):
        return a + b * x**c

    # A 3-parameter exponential function (a dependence function).
    def _exp3(x, a=0.0400, b=0.1748, c=-0.2243):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]
    power3 = DependenceFunction(_power3, bounds)
    exp3 = DependenceFunction(_exp3, bounds)

    dist_description_0 = {
        "distribution": WeibullDistribution(alpha=2.776, beta=1.471, gamma=0.8888),
    }
    dist_description_1 = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }
    model = GlobalHierarchicalModel([dist_description_0, dist_description_1])

    return model


def test_IFORMContour(seastate_model):
    """
    Compare the coordinates of an IFORM contour with the results from Haselsteiner
    et al. (2017; DOI: 10.1016/j.coastaleng.2017.03.002)
    """
    alpha = calculate_alpha(3, 25)
    my_contour = IFORMContour(seastate_model, alpha)

    my_coordinates = my_contour.coordinates
    np.testing.assert_allclose(max(my_coordinates[:, 0]), 15.23, atol=0.05)
    np.testing.assert_allclose(max(my_coordinates[:, 1]), 13.96, atol=0.05)

    np.testing.assert_raises(TypeError, IFORMContour, "string", alpha)


def test_ISORMContour(seastate_model):
    """
    Compare the coordinates of an ISORM contour with the results from Haselsteiner
    et al. (2017; DOI: 10.1016/j.coastaleng.2017.03.002, Fig. 8). The shown
    308.8-year IFORM contour is the same as an 25-year IFORM contour.
    """
    alpha = calculate_alpha(3, 25)
    my_contour = ISORMContour(seastate_model, alpha)

    my_coordinates = my_contour.coordinates
    np.testing.assert_allclose(max(my_coordinates[:, 0]), 17.4, atol=0.1)
    np.testing.assert_allclose(max(my_coordinates[:, 1]), 14.9, atol=0.1)


def test_DirectSamplingContour(seastate_model):
    """
    Computes a direct sampling contour and compares it with results
    from Huseby et al. (2013; DOI: 10.1016/j.oceaneng.2012.12.034, Tab. 5).
    """
    ref_contour_hs_1 = [
        9.99,
        10.65,
        10.99,
        11.25,
        11.25,
        11.41,
        11.42,
        11.46,
        11.48,
        11.54,
        11.57,
        11.56,
        11.58,
        11.59,
        11.59,
        11.60,
        11.60,
        11.59,
        11.59,
        11.56,
        11.53,
        11.46,
        11.26,
        10.88,
        7.44,
        2.05,
    ]
    ref_contour_tz_1 = [
        12.34,
        12.34,
        12.31,
        12.25,
        12.25,
        12.18,
        12.17,
        12.15,
        12.13,
        12.06,
        12.02,
        12.03,
        12.00,
        11.96,
        11.95,
        11.86,
        11.84,
        11.77,
        11.76,
        11.67,
        11.60,
        11.47,
        11.20,
        10.77,
        7.68,
        3.76,
    ]

    alpha = calculate_alpha(6, 1)

    # Fix the random seed for consistency. We do not want tests to randomly fail.
    # This method may come with some problems, bot for now we keep it, see
    # https://towardsdatascience.com/stop-using-numpy-random-seed-581a9972805f
    np.random.seed(seed=42)
    my_ds_contour = DirectSamplingContour(seastate_model, alpha, deg_step=6)

    my_coordinates = my_ds_contour.coordinates

    np.testing.assert_allclose(my_coordinates[0:26, 0], ref_contour_hs_1, atol=0.75)
    np.testing.assert_allclose(my_coordinates[0:26, 1], ref_contour_tz_1, atol=0.75)


def test_HighestDensityContour(seastate_model):
    """
    Compare the coordinates of a HD contour with the results from Haselsteiner
    et al. (2017; DOI: 10.1016/j.coastaleng.2017.03.002, Fig. 5)
    """
    alpha = calculate_alpha(3, 25)
    limits = [(0, 20), (0, 18)]
    deltas = [0.05, 0.05]
    my_contour = HighestDensityContour(seastate_model, alpha, limits, deltas)

    my_coordinates = my_contour.coordinates
    np.testing.assert_allclose(max(my_coordinates[:, 0]), 16.79, atol=0.05)
    np.testing.assert_allclose(max(my_coordinates[:, 1]), 14.64, atol=0.05)


def test_AndContour_correct_coordinates(seastate_model):
    alpha = 0.01
    n = 10000
    sample = seastate_model.draw_sample(n)
    contour = AndContour(seastate_model, alpha, deg_step=1, sample=sample)
    coords = contour.coordinates

    # Significant wave height at an angle of 0 deg.
    np.testing.assert_allclose(coords[0, 0], 9, atol=0.7)

    # Zero-up-crossing period at an angle of 90 deg.
    np.testing.assert_allclose(coords[-2, 1], 11, atol=0.5)


def test_AndContour_too_small_sample(seastate_model):
    alpha = 0.01
    with pytest.warns(UserWarning):
        n = 80  # Not sufficient datapoints to find precise alpha-regions.
        sample = seastate_model.draw_sample(n)
        AndContour(seastate_model, alpha, deg_step=1, sample=sample, allowed_error=0.01)


def test_OrContour_correct_coordinates(seastate_model):
    alpha = 0.01
    n = 10000
    sample = seastate_model.draw_sample(n)
    contour = OrContour(seastate_model, alpha, deg_step=1, sample=sample)
    coords = contour.coordinates

    # Significant wave height at the lowest angle (close to 0 deg).
    np.testing.assert_allclose(coords[0, 1], 11, atol=0.5)

    # Zero-up-crossing period at the highest angle (close to 90 deg).
    np.testing.assert_allclose(coords[-4, 0], 8.5, atol=0.7)


def test_OrContour_too_small_sample(seastate_model):
    alpha = 0.01
    with pytest.warns(UserWarning):
        n = 80  # Not sufficient datapoints to find precise alpha-regions.
        sample = seastate_model.draw_sample(n)
        OrContour(seastate_model, alpha, deg_step=1, sample=sample, allowed_error=0.01)
